from .VitPose.model import ViTPose
from .VitPose.utils.top_down_eval import keypoints_from_heatmaps
from .VitPose.configs import ViTPose_base_coco_256x192, ViTPose_large_coco_256x192, ViTPose_huge_coco_256x192
from HITrack.utils import download_models

from torchvision.transforms import transforms
import numpy as np
import torch
import cv2
import os

vitpose_model_map = {
    'b': (ViTPose_base_coco_256x192, 'vitpose-b-multi-coco.pth', '1R-fL7l6IJ7PT4cYTn-WEZa4r35q250J3'),
    'l': (ViTPose_large_coco_256x192, 'vitpose-l-multi-coco.pth', '17w-Au7eVSDfh8U5JJ0x57_0rWIfBNNEL'),
    'h': (ViTPose_huge_coco_256x192, 'vitpose-h-multi-coco.pth', '1O4Zaamrac_0pV5ELgKo0lepPob3KCwVx')
}


class VITPOSE:
    def __init__(self, model_version='b', device='cuda'):
        if model_version not in vitpose_model_map.keys():
            raise AssertionError(f'Only {", ".join(vitpose_model_map.keys())} models exist')

        self.cfg, model_name, cid = vitpose_model_map[model_version]
        self.model_path = os.path.join(os.path.dirname(__file__), 'VitPose', 'models', model_name)

        if not os.path.exists(self.model_path):
            download_models(cid, self.model_path, f"{model_name.split('.')[0]}")

        self.model_img_size = self.cfg.data_cfg['image_size']
        self.device = torch.device(device)
        self.model = self.model_init()

    def predict(self, img):
        org_w, org_h = img.shape[:2][::-1]

        # preprocess
        resized = cv2.resize(img, self.model_img_size, interpolation=cv2.INTER_AREA)
        img_tensor = transforms.ToTensor()(resized).unsqueeze(0).to(self.device)
        #         img_tensor = transforms.Compose (
        #                         [transforms.ToPILImage(),
        #                         transforms.Resize(self.model_img_size[::-1]),
        #                          transforms.ToTensor()]
        #                             )(img).unsqueeze(0).to(self.device)

        # process
        with torch.no_grad():
            heatmaps = self.model(img_tensor).detach().cpu().numpy()

        # decoder
        points, prob = keypoints_from_heatmaps(heatmaps=heatmaps, center=np.array([[org_w // 2, org_h // 2]]),
                                               scale=np.array([[org_w, org_h]]), unbiased=True, use_udp=True)
        keypoints = np.concatenate([points[:, :, ::-1], prob], axis=2)[0]
        return keypoints

    def __call__(self, rgb, boxes):
        keypoints_list = []
        for x1, y1, x2, y2 in boxes.astype(int):
            crop = rgb[y1:y2, x1:x2]
            keypoints = self.predict(crop)
            keypoints[:, :2] = keypoints[:, :2][:, ::-1] + [x1, y1]
            keypoints_list.append(keypoints)
        return np.float64(keypoints_list)

    def model_init(self):
        model = ViTPose(self.cfg.model)
        model.load_state_dict(torch.load(self.model_path)['state_dict'])
        model.to(self.device).eval()
        return model



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from .. import constant_init, normal_init
from .. import pose_pck_accuracy, flip_back, resize
import torch.nn.functional as F
from .topdown_heatmap_base_head import TopdownHeatmapBaseHead


class TopdownHeatmapSimpleHead(TopdownHeatmapBaseHead):
    """Top-down heatmap simple head. paper ref: Bin Xiao et al. ``Simple
    Baselines for Human Pose Estimation and Tracking``.

    TopdownHeatmapSimpleHead is consisted of (>=0) number of deconv layers
    and a simple conv2d layer.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        num_deconv_layers (int): Number of deconv layers.
            num_deconv_layers should >= 0. Note that 0 means
            no deconv layers.
        num_deconv_filters (list|tuple): Number of filters.
            If num_deconv_layers > 0, the length of
        num_deconv_kernels (list|tuple): Kernel sizes.
        in_index (int|Sequence[int]): Input feature index. Default: 0
        input_transform (str|None): Transformation type of input features.
            Options: 'resize_concat', 'multiple_select', None.
            Default: None.

            - 'resize_concat': Multiple feature maps will be resized to the
                same size as the first one and then concat together.
                Usually used in FCN head of HRNet.
            - 'multiple_select': Multiple feature maps will be bundle into
                a list and passed into decode head.
            - None: Only one select feature map is allowed.
        align_corners (bool): align_corners argument of F.interpolate.
            Default: False.
        loss_keypoint (dict): Config for keypoint loss. Default: None.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 num_deconv_layers=3,
                 num_deconv_filters=(256, 256, 256),
                 num_deconv_kernels=(4, 4, 4),
                 extra=None,
                 in_index=0,
                 input_transform=None,
                 align_corners=False,
                 loss_keypoint=None,
                 train_cfg=None,
                 test_cfg=None,
                 upsample=0,):
        super().__init__()

        self.in_channels = in_channels
        self.loss = loss_keypoint
        self.upsample = upsample

        self.train_cfg = {} if train_cfg is None else train_cfg
        self.test_cfg = {} if test_cfg is None else test_cfg
        self.target_type = self.test_cfg.get('target_type', 'GaussianHeatmap')

        self._init_inputs(in_channels, in_index, input_transform)
        self.in_index = in_index
        self.align_corners = align_corners

        if extra is not None and not isinstance(extra, dict):
            raise TypeError('extra should be dict or None.')

        if num_deconv_layers > 0:
            self.deconv_layers = self._make_deconv_layer(
                num_deconv_layers,
                num_deconv_filters,
                num_deconv_kernels,
            )
        elif num_deconv_layers == 0:
            self.deconv_layers = nn.Identity()
        else:
            raise ValueError(
                f'num_deconv_layers ({num_deconv_layers}) should >= 0.')

        identity_final_layer = False
        if extra is not None and 'final_conv_kernel' in extra:
            assert extra['final_conv_kernel'] in [0, 1, 3]
            if extra['final_conv_kernel'] == 3:
                padding = 1
            elif extra['final_conv_kernel'] == 1:
                padding = 0
            else:
                # 0 for Identity mapping.
                identity_final_layer = True
            kernel_size = extra['final_conv_kernel']
        else:
            kernel_size = 1
            padding = 0

        if identity_final_layer:
            self.final_layer = nn.Identity()
        else:
            conv_channels = num_deconv_filters[
                -1] if num_deconv_layers > 0 else self.in_channels

            layers = []
            if extra is not None:
                num_conv_layers = extra.get('num_conv_layers', 0)
                num_conv_kernels = extra.get('num_conv_kernels',
                                             [1] * num_conv_layers)

                for i in range(num_conv_layers):
                    layers.append(
                        nn.Conv2d(in_channels=conv_channels,
                                  out_channels=conv_channels,
                                  kernel_size=num_conv_kernels[i],
                                  stride=1,
                                  padding=(num_conv_kernels[i] - 1) // 2)
                        )
                    layers.append(nn.BatchNorm2d(conv_channels))
                    layers.append(nn.ReLU(inplace=True))

            layers.append(
                nn.Conv2d(in_channels=conv_channels,
                          out_channels=out_channels,
                          kernel_size=kernel_size,
                          stride=1,
                          padding=padding)
                )

            if len(layers) > 1:
                self.final_layer = nn.Sequential(*layers)
            else:
                self.final_layer = layers[0]

    def get_loss(self, output, target, target_weight):
        """Calculate top-down keypoint loss.

        Note:
            - batch_size: N
            - num_keypoints: K
            - heatmaps height: H
            - heatmaps weight: W

        Args:
            output (torch.Tensor[N,K,H,W]): Output heatmaps.
            target (torch.Tensor[N,K,H,W]): Target heatmaps.
            target_weight (torch.Tensor[N,K,1]):
                Weights across different joint types.
        """

        losses = dict()

        assert not isinstance(self.loss, nn.Sequential)
        assert target.dim() == 4 and target_weight.dim() == 3
        losses['heatmap_loss'] = self.loss(output, target, target_weight)

        return losses

    def get_accuracy(self, output, target, target_weight):
        """Calculate accuracy for top-down keypoint loss.

        Note:
            - batch_size: N
            - num_keypoints: K
            - heatmaps height: H
            - heatmaps weight: W

        Args:
            output (torch.Tensor[N,K,H,W]): Output heatmaps.
            target (torch.Tensor[N,K,H,W]): Target heatmaps.
            target_weight (torch.Tensor[N,K,1]):
                Weights across different joint types.
        """

        accuracy = dict()

        if self.target_type == 'GaussianHeatmap':
            _, avg_acc, _ = pose_pck_accuracy(
                output.detach().cpu().numpy(),
                target.detach().cpu().numpy(),
                target_weight.detach().cpu().numpy().squeeze(-1) > 0)
            accuracy['acc_pose'] = float(avg_acc)

        return accuracy

    def forward(self, x):
        """Forward function."""
        x = self._transform_inputs(x)
        x = self.deconv_layers(x)
        x = self.final_layer(x)
        return x

    def inference_model(self, x, flip_pairs=None):
        """Inference function.

        Returns:
            output_heatmap (np.ndarray): Output heatmaps.

        Args:
            x (torch.Tensor[N,K,H,W]): Input features.
            flip_pairs (None | list[tuple]):
                Pairs of keypoints which are mirrored.
        """
        output = self.forward(x)

        if flip_pairs is not None:
            output_heatmap = flip_back(
                output.detach().cpu().numpy(),
                flip_pairs,
                target_type=self.target_type)
            # feature is not aligned, shift flipped heatmap for higher accuracy
            if self.test_cfg.get('shift_heatmap', False):
                output_heatmap[:, :, :, 1:] = output_heatmap[:, :, :, :-1]
        else:
            output_heatmap = output.detach().cpu().numpy()
        return output_heatmap

    def _init_inputs(self, in_channels, in_index, input_transform):
        """Check and initialize input transforms.

        The in_channels, in_index and input_transform must match.
        Specifically, when input_transform is None, only single feature map
        will be selected. So in_channels and in_index must be of type int.
        When input_transform is not None, in_channels and in_index must be
        list or tuple, with the same length.

        Args:
            in_channels (int|Sequence[int]): Input channels.
            in_index (int|Sequence[int]): Input feature index.
            input_transform (str|None): Transformation type of input features.
                Options: 'resize_concat', 'multiple_select', None.

                - 'resize_concat': Multiple feature maps will be resize to the
                    same size as first one and than concat together.
                    Usually used in FCN head of HRNet.
                - 'multiple_select': Multiple feature maps will be bundle into
                    a list and passed into decode head.
                - None: Only one select feature map is allowed.
        """

        if input_transform is not None:
            assert input_transform in ['resize_concat', 'multiple_select']
        self.input_transform = input_transform
        self.in_index = in_index
        if input_transform is not None:
            assert isinstance(in_channels, (list, tuple))
            assert isinstance(in_index, (list, tuple))
            assert len(in_channels) == len(in_index)
            if input_transform == 'resize_concat':
                self.in_channels = sum(in_channels)
            else:
                self.in_channels = in_channels
        else:
            assert isinstance(in_channels, int)
            assert isinstance(in_index, int)
            self.in_channels = in_channels

    def _transform_inputs(self, inputs):
        """Transform inputs for decoder.

        Args:
            inputs (list[Tensor] | Tensor): multi-level img features.

        Returns:
            Tensor: The transformed inputs
        """
        if not isinstance(inputs, list):
            if not isinstance(inputs, list):
                if self.upsample > 0:
                    inputs = resize(
                        input=F.relu(inputs),
                        scale_factor=self.upsample,
                        mode='bilinear',
                        align_corners=self.align_corners
                        )
            return inputs

        if self.input_transform == 'resize_concat':
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(
                    input=x,
                    size=inputs[0].shape[2:],
                    mode='bilinear',
                    align_corners=self.align_corners) for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == 'multiple_select':
            inputs = [inputs[i] for i in self.in_index]
        else:
            inputs = inputs[self.in_index]

        return inputs

    def _make_deconv_layer(self, num_layers, num_filters, num_kernels):
        """Make deconv layers."""
        if num_layers != len(num_filters):
            error_msg = f'num_layers({num_layers}) ' \
                        f'!= length of num_filters({len(num_filters)})'
            raise ValueError(error_msg)
        if num_layers != len(num_kernels):
            error_msg = f'num_layers({num_layers}) ' \
                        f'!= length of num_kernels({len(num_kernels)})'
            raise ValueError(error_msg)

        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = \
                self._get_deconv_cfg(num_kernels[i])

            planes = num_filters[i]
            layers.append(
                nn.ConvTranspose2d(in_channels=self.in_channels,
                                   out_channels=planes,
                                   kernel_size=kernel,
                                   stride=2,
                                   padding=padding,
                                   output_padding=output_padding,
                                   bias=False)
                )
            layers.append(nn.BatchNorm2d(planes))
            layers.append(nn.ReLU(inplace=True))
            self.in_channels = planes

        return nn.Sequential(*layers)

    def init_weights(self):
        """Initialize model weights."""
        for _, m in self.deconv_layers.named_modules():
            if isinstance(m, nn.ConvTranspose2d):
                normal_init(m, std=0.001)
            elif isinstance(m, nn.BatchNorm2d):
                constant_init(m, 1)
        for m in self.final_layer.modules():
            if isinstance(m, nn.Conv2d):
                normal_init(m, std=0.001, bias=0)
            elif isinstance(m, nn.BatchNorm2d):
                constant_init(m, 1)



from yolov7.helpers import load_model
import os


class YOLOv7:
    def __init__(self, model_name='yolov7', conf=0.5, iou=0.45, device='cuda'):
        self.model_path = os.path.join(os.path.dirname(__file__),
                                       'models', model_name + '.pt')

        model = load_model(self.model_path, device=device)
        model.conf, model.iou, model.classes = conf, iou, [0]  # (optional list) filter by class
        self.device = device
        self.model = model

    def __call__(self, rgb):
        boxes = self.model(rgb, size=1280).pred[0].cpu().detach().numpy()[:, :4]
        return boxes


import torch.nn as nn
from einops import rearrange
from .module.trans import Transformer as Transformer_encoder
from .module.trans_hypothesis import Transformer as Transformer_hypothesis

class Model(nn.Module):
    def __init__(self, args):
        super().__init__()

        ## MHG
        self.norm_1 = nn.LayerNorm(args.frames)
        self.norm_2 = nn.LayerNorm(args.frames)
        self.norm_3 = nn.LayerNorm(args.frames)

        self.Transformer_encoder_1 = Transformer_encoder(4, args.frames, args.frames*2, length=2*args.n_joints, h=9)
        self.Transformer_encoder_2 = Transformer_encoder(4, args.frames, args.frames*2, length=2*args.n_joints, h=9)
        self.Transformer_encoder_3 = Transformer_encoder(4, args.frames, args.frames*2, length=2*args.n_joints, h=9)

        ## Embedding
        if args.frames > 27:
            self.embedding_1 = nn.Conv1d(2*args.n_joints, args.channel, kernel_size=1)
            self.embedding_2 = nn.Conv1d(2*args.n_joints, args.channel, kernel_size=1)
            self.embedding_3 = nn.Conv1d(2*args.n_joints, args.channel, kernel_size=1)
        else:
            self.embedding_1 = nn.Sequential(
                nn.Conv1d(2*args.n_joints, args.channel, kernel_size=1),
                nn.BatchNorm1d(args.channel, momentum=0.1),
                nn.ReLU(inplace=True),
                nn.Dropout(0.25)
            )

            self.embedding_2 = nn.Sequential(
                nn.Conv1d(2*args.n_joints, args.channel, kernel_size=1),
                nn.BatchNorm1d(args.channel, momentum=0.1),
                nn.ReLU(inplace=True),
                nn.Dropout(0.25)
            )

            self.embedding_3 = nn.Sequential(
                nn.Conv1d(2*args.out_joints, args.channel, kernel_size=1),
                nn.BatchNorm1d(args.channel, momentum=0.1),
                nn.ReLU(inplace=True),
                nn.Dropout(0.25)
            )

        ## SHR & CHI
        self.Transformer_hypothesis = Transformer_hypothesis(args.layers, args.channel, args.d_hid, length=args.frames)
        
        ## Regression
        self.regression = nn.Sequential(
            nn.BatchNorm1d(args.channel*3, momentum=0.1),
            nn.Conv1d(args.channel*3, 3*args.out_joints, kernel_size=1)
        )

    def forward(self, x):
        B, F, J, C = x.shape
        x = rearrange(x, 'b f j c -> b (j c) f').contiguous()

        ## MHG
        x_1 = x   + self.Transformer_encoder_1(self.norm_1(x))
        x_2 = x_1 + self.Transformer_encoder_2(self.norm_2(x_1)) 
        x_3 = x_2 + self.Transformer_encoder_3(self.norm_3(x_2))
        
        ## Embedding
        x_1 = self.embedding_1(x_1).permute(0, 2, 1).contiguous() 
        x_2 = self.embedding_2(x_2).permute(0, 2, 1).contiguous()
        x_3 = self.embedding_3(x_3).permute(0, 2, 1).contiguous()

        ## SHR & CHI
        x = self.Transformer_hypothesis(x_1, x_2, x_3) 

        ## Regression
        x = x.permute(0, 2, 1).contiguous() 
        x = self.regression(x) 
        x = rearrange(x, 'b (j c) f -> b f j c', j=J).contiguous()

        return x









from scipy.signal import savgol_filter
from tqdm.notebook import tqdm
from itertools import groupby
import numpy as np
import requests
import warnings
import torch
import cv2
import gc
import os
warnings.filterwarnings("ignore")


def download_models(cid, path='', title=None):
    if len(os.path.basename(path).split('.')) == 2:
        path = os.path.dirname(path)

    url = f'https://drive.google.com/uc?export=download&confirm=no_antivirus&id={cid}'
    with requests.get(url, stream=True) as response:
        total_size = int(response.headers["Content-Length"])
        filename = response.headers['Content-Disposition'].split('"')[1]
        pbar = tqdm(response.iter_content(1024), title,
                    total=total_size, unit="B", unit_scale=True, unit_divisor=1024)
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, filename), "wb") as f:
            for data in pbar.iterable:
                f.write(data)
                pbar.update(len(data))
    pbar.close()


def clear():
    torch.cuda.empty_cache()
    gc.collect()


def id_existence(keypoints):
    return ~(keypoints == 0).all(2).all(2)


def recover_kps(keypoints, max_=10):
    keypoints = keypoints.copy()
    exist = ~(keypoints == 0).all(2).all(2)
    for id_ in range(len(keypoints)):
        rep = [[k, len(list(g))] for k, g in groupby(exist[id_])]
        starts = np.array([0] + [x[1] for x in rep]).cumsum()

        for i, (exs, val) in enumerate(rep[1:-1], start=1):
            if not exs and val < max_:
                left = keypoints[id_][starts[i] - 1]
                right = keypoints[id_][starts[i + 1]]

                for k in range(val):
                    alpha = (k + 1) / (val + 1)
                    keypoints[id_][starts[i] + k] = left * (1 - alpha) + right * alpha

                print(f'id:{id_}, frames recovered: {starts[i]}-{starts[i + 1]}')
    return keypoints


def merge_track(keypoints, merge_dict: dict):
    exist = ~(keypoints == 0).all(2).all(2)
    keypoints = keypoints.copy()
    for old, new in merge_dict.items():
        keypoints[old][exist[new]] = keypoints[new][exist[new]]
    keypoints = np.delete(keypoints, list(merge_dict.values()), axis=0)
    return keypoints


def world2cam(pose, pose3d, cameraMatrix, distCoeffs, rvec=None, tvec=None):
    use = False if rvec is None else True
    rvec, tvec = cv2.solvePnP(pose3d, pose, cameraMatrix, distCoeffs, rvec, tvec, use)[1:]
    R = cv2.Rodrigues(rvec)[0]
    x, y, z = R @ pose3d.T + tvec
    x, y, z = x, z, -y
    return np.array((x, y, z)).T, rvec, tvec


def CreateCamera(img_size, focus=1):
    fx = fy = focus * np.hypot(*img_size)
    cx, cy = img_size[0] / 2, img_size[1] / 2

    distCoeffs = np.zeros(4, np.float32)
    cameraMatrix = np.float32([[fx, 0, cx],
                               [0, fy, cy],
                               [0, 0, 1]])
    return cameraMatrix, distCoeffs


def poses2scene(data, img_size=(1920, 1080), smooth=True):
    exist = data.ids != -1
    scene = np.zeros_like(data.keypoints_3d)
    cameraMatrix, distCoeffs = CreateCamera(img_size)
    
    lis = []
    rvec, tvec = None, None
    for pose, pose3d in zip(coco2h36(data.keypoints_2d)[exist], data.keypoints_3d[exist]):
        sc, rvec, tvec = world2cam(pose, pose3d, cameraMatrix, distCoeffs, rvec, tvec) 
        lis.append(sc)
        
    scene[exist] = lis

    if smooth:
        scene = smoothing(scene)

    return scene


def smoothing(scene, window=11, poly=5):
    scene = scene.copy()
    n_ids, n_frames, n_keypoints, n_dims = scene.shape
    for id_ in range(n_ids):
        for point in range(n_keypoints):
            for dim in range(n_dims):
                scene[id_, :, point, dim] = savgol_filter(
                    scene[id_, :, point, dim], window, poly)
    return scene


class VideoDataKeypoints:

    def __init__(self, path='test.npz', keypoints_2d=None, keypoints_3d=None, scene=None):
        self.keypoints_2d, self.keypoints_3d, self.scene = self.check([keypoints_2d, keypoints_3d, scene])
        self.shape, self.ids = self.create_ids()
        self.path = path

    def save(self):
        kps2d, kps3d, scne = [arr[self.ids != -1] if arr is not None else None
                              for arr in [self.keypoints_2d, self.keypoints_3d, self.scene]]

        np.savez_compressed(self.path, keypoints_2d=kps2d, keypoints_3d=kps3d, scene=scne, ids=self.ids,
                            shape=self.shape)

    def load(self):
        with np.load(self.path, allow_pickle=True) as npz:
            self.ids, self.shape = npz['ids'], npz['shape']

            arrays = []
            for name, dims in zip(npz.files[:3], [2, 3, 3]):
                array = None
                if npz[name].ndim:
                    array = np.zeros((*self.shape, dims))
                    array[self.ids != -1] = npz[name]
                arrays.append(array)

        self.keypoints_2d, self.keypoints_3d, self.scene = arrays

    def update(self, keypoints_2d=None, keypoints_3d=None, scene=None, save=False):
        arrays = self.check([keypoints_2d, keypoints_3d, scene])
        if keypoints_2d is not None:
            self.keypoints_2d, self.keypoints_3d, self.scene = arrays
        else:
            if keypoints_3d is not None:
                self.keypoints_3d = arrays[1]
            if scene is not None:
                self.scene = arrays[2]

        self.shape, self.ids = self.create_ids()
        if save:
            self.save()

    def create_ids(self):
        for arr in [self.keypoints_2d, self.keypoints_3d, self.scene]:
            if arr is not None:
                n_ids, n_frames, _, _ = arr.shape
                ids = np.mgrid[:n_ids, :n_frames][0]
                ids[~id_existence(arr)] = -1
                return arr.shape[:3], ids
        return None, None

    @staticmethod
    def check(arrays):
        shapes = [arr.shape for arr in arrays if arr is not None]
        if len(shapes) > 0:
            assert all([len(shape) == 4 for shape in shapes]), 'All arrays must be 4-dimensional'
            assert np.all(np.diff(np.array(shapes)[:, :3], axis=0) == 0), 'Shapes of arrays must be the same'
            assert np.all(np.diff([id_existence(arr) for arr in arrays if arr is not None],
                                  axis=0) == 0), 'Arrays do not match by the number of ids on some frames'

            arrays = [np.float64(arr) if arr is not None else None for arr in arrays]
        return arrays


class OpenVideo:
    def __init__(self, video_path, end=None, pbar=True):
        self.cap = cv2.VideoCapture(video_path)
        self.shape = int(self.cap.get(3)), int(self.cap.get(4))
        self.fps = self.cap.get(5)
        self.length = end if end else int(self.cap.get(7))
        self.pbar = pbar
        self.frame = 0
        
        if pbar:
            self.tqdm = tqdm(total=self.length)

    def read(self):
        ret, bgr = self.cap.read()
        self.frame += 1
        if self.pbar:
            self.tqdm.update(1)
            
        if self.frame == self.length:
            self.cap.release()
            if self.pbar:
                self.tqdm.close()

        if ret:
            rgb = cv2.cvtColor(bgr, cv2.COLOR_BGR2RGB)
            return rgb
        else:
            self.cap.release()
            if self.pbar:
                self.tqdm.close()


def coco2h36(keypoints):
    # shape (ids, frames, 17, 2)
    h36m_coco_order = [9, 11, 14, 12, 15, 13, 16, 4, 1, 5, 2, 6, 3]
    coco_order = [0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
    spple_keypoints = [10, 8, 0, 7]

    keypoints_h36m = np.zeros_like(keypoints, dtype=np.float64)
    htps_keypoints = np.zeros((*keypoints.shape[:2], 4, 2), dtype=np.float64)

    # htps_keypoints: head, thorax, pelvis, spine
    htps_keypoints[:, :, 0, 0] = np.mean(keypoints[:, :, 1:5, 0], axis=2)
    htps_keypoints[:, :, 0, 1] = np.sum(keypoints[:, :, 1:3, 1], axis=2) - keypoints[:, :, 0, 1]
    htps_keypoints[:, :, 1, :] = np.mean(keypoints[:, :, 5:7, :], axis=2)
    htps_keypoints[:, :, 1, :] += (keypoints[:, :, 0, :] - htps_keypoints[:, :, 1, :]) / 3

    htps_keypoints[:, :, 2, :] = np.mean(keypoints[:, :, 11:13, :], axis=2)
    htps_keypoints[:, :, 3, :] = np.mean(keypoints[:, :, [5, 6, 11, 12], :], axis=2)

    keypoints_h36m[:, :, spple_keypoints, :] = htps_keypoints
    keypoints_h36m[:, :, h36m_coco_order, :] = keypoints[:, :, coco_order, :]

    keypoints_h36m[:, :, 9, :] -= (keypoints_h36m[:, :, 9, :] - np.mean(keypoints[:, :, 5:7, :], axis=2)) / 4
    keypoints_h36m[:, :, 7, 0] += 0.3 * (keypoints_h36m[:, :, 7, 0] - np.mean(keypoints_h36m[:, :, [0, 8], 0], axis=2))
    keypoints_h36m[:, :, 8, 1] -= (np.mean(keypoints[:, :, 1:3, 1], axis=2) - keypoints[:, :, 0, 1]) * 2 / 3

    return keypoints_h36m



import torch
import torch.nn as nn
from functools import partial
from timm.models.layers import DropPath

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

class Cross_Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.linear_q = nn.Linear(dim, dim, bias=qkv_bias)
        self.linear_k = nn.Linear(dim, dim, bias=qkv_bias)
        self.linear_v = nn.Linear(dim, dim, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x_1, x_2, x_3):
        B, N, C = x_1.shape
        q = self.linear_q(x_1).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        k = self.linear_k(x_2).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        v = self.linear_v(x_3).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

class SHR_Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_hidden_dim, qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1_1 = norm_layer(dim)
        self.norm1_2 = norm_layer(dim)
        self.norm1_3 = norm_layer(dim)

        self.attn_1 = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.attn_2 = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.attn_3 = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim * 3)
        self.mlp = Mlp(in_features=dim * 3, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x_1, x_2, x_3):
        x_1 = x_1 + self.drop_path(self.attn_1(self.norm1_1(x_1)))
        x_2 = x_2 + self.drop_path(self.attn_2(self.norm1_2(x_2)))
        x_3 = x_3 + self.drop_path(self.attn_3(self.norm1_3(x_3)))

        x = torch.cat([x_1, x_2, x_3], dim=2)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        x_1 = x[:, :, :x.shape[2] // 3]
        x_2 = x[:, :, x.shape[2] // 3: x.shape[2] // 3 * 2]
        x_3 = x[:, :, x.shape[2] // 3 * 2: x.shape[2]]

        return  x_1, x_2, x_3

class CHI_Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_hidden_dim, qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm3_11 = norm_layer(dim)
        self.norm3_12 = norm_layer(dim)
        self.norm3_13 = norm_layer(dim)

        self.norm3_21 = norm_layer(dim)
        self.norm3_22 = norm_layer(dim)
        self.norm3_23 = norm_layer(dim)

        self.norm3_31 = norm_layer(dim)
        self.norm3_32 = norm_layer(dim)
        self.norm3_33 = norm_layer(dim)

        self.attn_1 = Cross_Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.attn_2 = Cross_Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.attn_3 = Cross_Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim * 3)
        self.mlp = Mlp(in_features=dim * 3, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x_1, x_2, x_3):
        x_1 = x_1 + self.drop_path(self.attn_1(self.norm3_11(x_2), self.norm3_12(x_3), self.norm3_13(x_1)))    
        x_2 = x_2 + self.drop_path(self.attn_2(self.norm3_21(x_1), self.norm3_22(x_3), self.norm3_23(x_2)))  
        x_3 = x_3 + self.drop_path(self.attn_3(self.norm3_31(x_1), self.norm3_32(x_2), self.norm3_33(x_3)))  

        x = torch.cat([x_1, x_2, x_3], dim=2)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        x_1 = x[:, :, :x.shape[2] // 3]
        x_2 = x[:, :, x.shape[2] // 3: x.shape[2] // 3 * 2]
        x_3 = x[:, :, x.shape[2] // 3 * 2: x.shape[2]]

        return  x_1, x_2, x_3

class Transformer(nn.Module):
    def __init__(self, depth=3, embed_dim=512, mlp_hidden_dim=1024, h=8, drop_rate=0.1, length=27):
        super().__init__()
        drop_path_rate = 0.20
        attn_drop_rate = 0.
        qkv_bias = True
        qk_scale = None

        norm_layer = partial(nn.LayerNorm, eps=1e-6)

        self.pos_embed_1 = nn.Parameter(torch.zeros(1, length, embed_dim))
        self.pos_embed_2 = nn.Parameter(torch.zeros(1, length, embed_dim))
        self.pos_embed_3 = nn.Parameter(torch.zeros(1, length, embed_dim))

        self.pos_drop_1 = nn.Dropout(p=drop_rate)
        self.pos_drop_2 = nn.Dropout(p=drop_rate)
        self.pos_drop_3 = nn.Dropout(p=drop_rate)

        # stochastic depth decay rule
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  

        self.SHR_blocks = nn.ModuleList([
            SHR_Block(
                dim=embed_dim, num_heads=h, mlp_hidden_dim=mlp_hidden_dim, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth-1)])

        self.CHI_blocks = nn.ModuleList([
            CHI_Block(
                dim=embed_dim, num_heads=h, mlp_hidden_dim=mlp_hidden_dim, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[depth-1], norm_layer=norm_layer)
            for i in range(1)])

        self.norm = norm_layer(embed_dim * 3)

    def forward(self, x_1, x_2, x_3):
        x_1 += self.pos_embed_1
        x_2 += self.pos_embed_2
        x_3 += self.pos_embed_3

        x_1 = self.pos_drop_1(x_1)
        x_2 = self.pos_drop_2(x_2)
        x_3 = self.pos_drop_3(x_3)

        for i, blk in enumerate(self.SHR_blocks):
            x_1, x_2, x_3 = self.SHR_blocks[i](x_1, x_2, x_3)

        x_1, x_2, x_3 = self.CHI_blocks[0](x_1, x_2, x_3)

        x = torch.cat([x_1, x_2, x_3], dim=2)
        x = self.norm(x)

        return x






import os
import warnings
import random
import numpy as np

from collections import OrderedDict
# import os.path as osp

import torch
import torch.nn as nn

from torch import distributed as dist
from torch.nn.parallel import DataParallel, DistributedDataParallel
from typing import Tuple
# from .dist_util import get_dist_info

MODULE_WRAPPERS = [DataParallel, DistributedDataParallel]


def get_dist_info() -> Tuple[int, int]:
    if dist.is_available() and dist.is_initialized():
        rank = dist.get_rank()
        world_size = dist.get_world_size()
    else:
        rank = 0
        world_size = 1
    return rank, world_size


def init_random_seed(seed=None, device='cuda'):
    """Initialize random seed.

    If the seed is not set, the seed will be automatically randomized,
    and then broadcast to all processes to prevent some potential bugs.

    Args:
        seed (int, Optional): The seed. Default to None.
        device (str): The device where the seed will be put on.
            Default to 'cuda'.

    Returns:
        int: Seed to be used.
    """
    if seed is not None:
        return seed

    # Make sure all ranks share the same random seed to prevent
    # some potential bugs. Please refer to
    # https://github.com/open-mmlab/mmdetection/issues/6339
    rank, world_size = get_dist_info()
    seed = np.random.randint(2**31)
    if world_size == 1:
        return seed

    if rank == 0:
        random_num = torch.tensor(seed, dtype=torch.int32, device=device)
    else:
        random_num = torch.tensor(0, dtype=torch.int32, device=device)
    dist.broadcast(random_num, src=0)
    return random_num.item()


def set_random_seed(seed: int,
                    deterministic: bool = False,
                    use_rank_shift: bool = False) -> None:
    """Set random seed.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
        rank_shift (bool): Whether to add rank number to the random seed to
            have different random seed in different threads. Default: False.
    """
    if use_rank_shift:
        rank, _ = get_dist_info()
        seed += rank
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

def is_module_wrapper(module: nn.Module) -> bool:
    """ Check if module wrrapper exists recursively """
    def is_module_in_wrapper(module, module_wrapper):
        module_wrappers = tuple(module_wrapper.module_dict.values())
        if isinstance(module, module_wrappers):
            return True
        for child in module_wrapper.children.values():
            if is_module_in_wrapper(module, child):
                return True
    return is_module_in_wrapper(module, MODULE_WRAPPERS)


def load_state_dict(module, state_dict, strict=False, logger=None):
    """Load state_dict to a module.

    This method is modified from :meth:`torch.nn.Module.load_state_dict`.
    Default value for ``strict`` is set to ``False`` and the message for
    param mismatch will be shown even if strict is False.

    Args:
        module (Module): Module that receives the state_dict.
        state_dict (OrderedDict): Weights.
        strict (bool): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``False``.
        logger (:obj:`logging.Logger`, optional): Logger to log the error
            message. If not specified, print function will be used.
    """
    unexpected_keys = []
    all_missing_keys = []
    err_msg = []

    metadata = getattr(state_dict, '_metadata', None)
    state_dict = state_dict.copy()
    if metadata is not None:
        state_dict._metadata = metadata

    # use _load_from_state_dict to enable checkpoint version control
    def load(module, prefix=''):
        # recursively check parallel module in case that the model has a
        # complicated structure, e.g., nn.Module(nn.Module(DDP))
        if is_module_wrapper(module):
            module = module.module
        local_metadata = {} if metadata is None else metadata.get(
            prefix[:-1], {})
        module._load_from_state_dict(state_dict, prefix, local_metadata, True,
                                     all_missing_keys, unexpected_keys,
                                     err_msg)
        for name, child in module._modules.items():
            if child is not None:
                load(child, prefix + name + '.')

    load(module)
    load = None  # break load->load reference cycle

    # ignore "num_batches_tracked" of BN layers
    missing_keys = [
        key for key in all_missing_keys if 'num_batches_tracked' not in key
    ]

    if unexpected_keys:
        err_msg.append('unexpected key in source '
                       f'state_dict: {", ".join(unexpected_keys)}\n')
    if missing_keys:
        err_msg.append(
            f'missing keys in source state_dict: {", ".join(missing_keys)}\n')

    rank, _ = get_dist_info()
    if len(err_msg) > 0 and rank == 0:
        err_msg.insert(
            0, 'The model and loaded state dict do not match exactly\n')
        err_msg = '\n'.join(err_msg)
        if strict:
            raise RuntimeError(err_msg)
        elif logger is not None:
            logger.warning(err_msg)
        else:
            print(err_msg)


def load_checkpoint(model,
                    filename,
                    map_location='cpu',
                    strict=False,
                    logger=None):
    """Load checkpoint from a file or URI.

    Args:
        model (Module): Module to load checkpoint.
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``.
        map_location (str): Same as :func:`torch.load`.
        strict (bool): Whether to allow different params for the model and
            checkpoint.
        logger (:mod:`logging.Logger` or None): The logger for error message.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    checkpoint = torch.load(filename, map_location=map_location)
    # OrderedDict is a subclass of dict
    if not isinstance(checkpoint, dict):
        raise RuntimeError(
            f'No state_dict found in checkpoint file {filename}')
    # get state_dict from checkpoint
    if 'state_dict' in checkpoint:
        state_dict_tmp = checkpoint['state_dict']
    else:
        state_dict_tmp = checkpoint

    state_dict = OrderedDict()
    # strip prefix of state_dict
    for k, v in state_dict_tmp.items():
        if k.startswith('module.backbone.'):
            state_dict[k[16:]] = v
        elif k.startswith('module.'):
            state_dict[k[7:]] = v
        elif k.startswith('backbone.'):
            state_dict[k[9:]] = v
        else:
            state_dict[k] = v
    # load state_dict
    load_state_dict(model, state_dict, strict, logger)
    return checkpoint


def resize(input,
           size=None,
           scale_factor=None,
           mode='nearest',
           align_corners=None,
           warning=True):
    if warning:
        if size is not None and align_corners:
            input_h, input_w = tuple(int(x) for x in input.shape[2:])
            output_h, output_w = tuple(int(x) for x in size)
            if output_h > input_h or output_w > output_h:
                if ((output_h > 1 and output_w > 1 and input_h > 1
                     and input_w > 1) and (output_h - 1) % (input_h - 1)
                        and (output_w - 1) % (input_w - 1)):
                    warnings.warn(
                        f'When align_corners={align_corners}, '
                        'the output would more aligned if '
                        f'input size {(input_h, input_w)} is `x+1` and '
                        f'out size {(output_h, output_w)} is `nx+1`')
    if isinstance(size, torch.Size):
        size = tuple(int(x) for x in size)
        
def constant_init(module: nn.Module, val: float, bias: float = 0) -> None:
    if hasattr(module, 'weight') and module.weight is not None:
        nn.init.constant_(module.weight, val)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)

def normal_init(module: nn.Module,
                mean: float = 0,
                std: float = 1,
                bias: float = 0) -> None:
    if hasattr(module, 'weight') and module.weight is not None:
        nn.init.normal_(module.weight, mean, std)
    if hasattr(module, 'bias') and module.bias is not None:
        nn.init.constant_(module.bias, bias)


import torch
import torch.nn as nn
from functools import partial
from timm.models.layers import DropPath

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  
        
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

class Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_hidden_dim, qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, \
            qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x

class Transformer(nn.Module):
    def __init__(self, depth=3, embed_dim=512, mlp_hidden_dim=1024, h=8, drop_rate=0.1, length=27):
        super().__init__()
        drop_path_rate = 0.2
        attn_drop_rate = 0.
        qkv_bias = True
        qk_scale = None

        norm_layer = partial(nn.LayerNorm, eps=1e-6)

        self.pos_embed = nn.Parameter(torch.zeros(1, length, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  

        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=h, mlp_hidden_dim=mlp_hidden_dim, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth)])

        self.norm = norm_layer(embed_dim)

    def forward(self, x):
        x += self.pos_embed
        x = self.pos_drop(x)

        for blk in self.blocks:
            x = blk(x)

        x = self.norm(x)

        return x







from ..constants import H36_PERSON_SKELETON, COLORS
from . import HumanVisualizer
from ..utils import coco2h36, OpenVideo
import cv2
import numpy as np
from tqdm.notebook import tqdm
import pandas as pd
import plotly.express as px
import io
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt

plt.switch_backend('agg')


class Visualizer:
    def __init__(self, data, video_path=None, skeleton_format='COCO'):
        self.data = data
        self.video_path = video_path
        self.size = (1080, 1080)
        self.fps = 30
        self.pairs = np.array(H36_PERSON_SKELETON) - 1
        self.skeleton_format = skeleton_format

    def __call__(self, how='3D_scene', id_=None, end=None, compress=False, original_sound=False):
        self.video_out = f'{self.video_path}.{how}.mp4'
        self.len_video = self.data.shape[1] if end is None else int(end)

        if how == '3D_scene':
            flat = self.data.scene[self.data.ids != -1].reshape(-1, 3)
            range_axes = np.concatenate((flat.min(0)[None], flat.max(0)[None])).T
            scale_axes = np.diff(range_axes)[:, 0]

        elif how == '3D_single':
            range_axes = [[-0.7, 0.7] for _ in range(3)]
            scale_axes = [1, 1, 1]
            if id_ is None:
                raise AssertionError('Specify id_')

        else:
            assert how in ['2D', '3D_single', '3D_scene']

        if self.video_path:
            cap = OpenVideo(self.video_path, self.len_video, False)
            self.fps = cap.fps
            w, h = cap.shape
            self.size = (w, h) if how == '2D' else (w + h, h)

            vis_2d = HumanVisualizer(skeleton_format=self.skeleton_format)
            keypoints_2d = self.data.keypoints_2d
            if self.skeleton_format == 'H36':
                keypoints_2d = coco2h36(keypoints_2d)
        
        fig = plt.figure(figsize=(10, 10), dpi=self.size[1] / 10)
        result = cv2.VideoWriter(self.video_out, cv2.VideoWriter_fourcc(*'MP4V'), self.fps, self.size)
        for i in tqdm(range(self.len_video)):

            if how == '3D_scene':
                ids = self.data.ids[:, i]
                poses = self.data.scene[:, i]

            elif how == '3D_single':
                ids = [id_]
                poses = self.data.keypoints_3d[ids, i]

            if how != '2D':
                image_3d = self.matplotlib_3D(fig, poses, ids, range_axes, scale_axes)

            if self.video_path:
                rgb = cap.read()
                rgb = vis_2d(rgb, keypoints_2d[:, i], self.data.ids[:, i])
                if how != '2D':
                    image = np.concatenate((rgb, image_3d), axis=1)
                else:
                    image = rgb
            else:
                image = image_3d

            result.write(image[:, :, ::-1])
        result.release()

        if compress or original_sound:
            print('ffmpeg starting...')
            self.ffmpeg(compress, original_sound)
            print('Done')

    def matplotlib_3D(self, fig, poses, ids, range_axes, scale_axes):
        ax = Axes3D(fig, elev=15, azim=80, auto_add_to_figure=False)
        fig.add_axes(ax)

        for id_, pose in zip(ids, poses):
            if id_ != -1:
                for pair in pose[self.pairs]:
                    x, y, z = pair.T
                    ax.plot(x, y, z, lw=2, color=np.array(COLORS[id_]) / 255)

        ax.set_xlim3d(range_axes[0][::-1])
        ax.set_ylim3d(range_axes[1][::-1])
        ax.set_zlim3d(range_axes[2])
        ax.set_box_aspect(scale_axes)

        white = (1.0, 1.0, 1.0, 0.0)
        ax.xaxis.set_pane_color(white)
        ax.yaxis.set_pane_color(white)
        ax.zaxis.set_pane_color(white)

        ax.tick_params('x', labelbottom=False)
        ax.tick_params('y', labelleft=False)
        ax.tick_params('z', labelleft=False)

        image = save_ax_nosave()
        fig.clf()

        return image

    def ffmpeg(self, compress=True, original_sound=True, replace=True):
        # ffmpeg required
        import subprocess, os
        if original_sound:
            subprocess.call(f'ffmpeg -i {self.video_path} -q:a 0 -map a temp.mp3', shell=True)

            cmd = f"ffmpeg -i {self.video_out} -i temp.mp3 -c copy -map 0:v:0 -map 1:a:0 temp.mp4"
            subprocess.call(cmd, shell=True)

            os.remove('temp.mp3')
            os.remove(self.video_out)
            os.rename('temp.mp4', self.video_out)

        if compress:
            subprocess.call('ffmpeg -i {} -acodec libmp3lame -ab 192 {} -y'.format(self.video_out,
                                                    self.video_out + '.mp4'), stderr=subprocess.DEVNULL, shell=True)
            if replace:
                os.remove(self.video_out)
                os.rename(self.video_out + '.mp4', self.video_out)


def save_ax_nosave():
    buff = io.BytesIO()
    plt.savefig(buff, format="png", bbox_inches='tight', pad_inches=0)
    buff.seek(0)
    im = plt.imread(buff)
    return np.uint8(im[:, :, :3] * 255)


def scene_plotly(scene, thickness=4, show_axes=False):
    assert len(scene.shape) == 4
    exist = ~(scene == 0).all(2).all(2)
    n_ids, n_frames = exist.shape
    pairs = np.array(H36_PERSON_SKELETON) - 1
    Nones = np.full((n_ids, n_frames, len(pairs), 1, 3), None)
    px_poses = np.concatenate((scene[:, :, pairs], Nones), 3)
    px_poses = px_poses[exist].reshape(-1, 3)

    ids_frames = np.tile(np.mgrid[0:n_ids, 0:n_frames][:, exist].T, len(pairs) * 3).reshape(-1, 2)
    df = pd.DataFrame(np.concatenate((px_poses, ids_frames), 1),
                      columns=['x', 'y', 'z', 'id_', 'frame']).apply(pd.to_numeric)

    fig = px.line_3d(df, x='x', y='y', z='z', color='id_', animation_frame='frame',
                     color_discrete_sequence=rgb2hex(COLORS))

    min_max = np.concatenate((df.min()[:3].values[None], df.max()[:3].values[None])).T
    diff = np.diff(min_max)[:, 0]
    scale = diff.mean()
    title = None if show_axes else ''

    scene_aspectratio = dict(x=diff[0], y=diff[1], z=diff[2])
    scene = dict(xaxis=dict(nticks=5, range=min_max[0], showticklabels=show_axes, title=title),
                 yaxis=dict(nticks=5, range=min_max[1], showticklabels=show_axes, title=title),
                 zaxis=dict(nticks=5, range=min_max[2], showticklabels=show_axes, title=title))

    camera = dict(up=dict(x=0, y=0, z=1),
                  center=dict(x=0, y=0, z=0),
                  eye=dict(x=-0.5 * scale, y=-1.7 * scale, z=0.6 * scale))

    fig.update_layout(showlegend=False,
                      width=900, height=500, margin=dict(t=10, r=0, l=0, b=0), scene=scene,
                      scene_camera=camera, scene_aspectmode='manual', scene_aspectratio=scene_aspectratio,
                      )

    fig.update_traces(line_width=thickness)
    return fig


def rgb2hex(colors):
    return ["".join([hex(rgb)[2:] for rgb in c]) for c in colors]



# Copyright (c) OpenMMLab. All rights reserved.
import warnings

import cv2
import numpy as np

from .post_processing import transform_preds


def heatmap2coords(heatmaps: np.ndarray, original_resolution: tuple[int, int]=(256, 192)) -> np.ndarray:
    __, __, heatmap_h, heatmap_w = heatmaps.shape
    output = []
    for heatmap in heatmaps:
        keypoint_coords = []
        for joint in heatmap:
            keypoint_coord = np.unravel_index(np.argmax(joint), (heatmap_h, heatmap_w))
            """
            - 0: coord_y / (height//4) * bbox_height + bb_y1
            - 1: coord_x / (width//4) * bbox_width + bb_x1
            - 2: confidences
            """
            coord_y = keypoint_coord[0] / heatmap_h*original_resolution[0]
            coord_x = keypoint_coord[1] / heatmap_w*original_resolution[1]
            prob = joint[keypoint_coord]
            keypoint_coords.append([coord_y, coord_x, prob])
        output.append(keypoint_coords)
            
    return np.array(output).astype(float)

def _calc_distances(preds, targets, mask, normalize):
    """Calculate the normalized distances between preds and target.

    Note:
        batch_size: N
        num_keypoints: K
        dimension of keypoints: D (normally, D=2 or D=3)

    Args:
        preds (np.ndarray[N, K, D]): Predicted keypoint location.
        targets (np.ndarray[N, K, D]): Groundtruth keypoint location.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.
        normalize (np.ndarray[N, D]): Typical value is heatmap_size

    Returns:
        np.ndarray[K, N]: The normalized distances. \
            If target keypoints are missing, the distance is -1.
    """
    N, K, _ = preds.shape
    # set mask=0 when normalize==0
    _mask = mask.copy()
    _mask[np.where((normalize == 0).sum(1))[0], :] = False
    distances = np.full((N, K), -1, dtype=np.float32)
    # handle invalid values
    normalize[np.where(normalize <= 0)] = 1e6
    distances[_mask] = np.linalg.norm(
        ((preds - targets) / normalize[:, None, :])[_mask], axis=-1)
    return distances.T


def _distance_acc(distances, thr=0.5):
    """Return the percentage below the distance threshold, while ignoring
    distances values with -1.

    Note:
        batch_size: N
    Args:
        distances (np.ndarray[N, ]): The normalized distances.
        thr (float): Threshold of the distances.

    Returns:
        float: Percentage of distances below the threshold. \
            If all target keypoints are missing, return -1.
    """
    distance_valid = distances != -1
    num_distance_valid = distance_valid.sum()
    if num_distance_valid > 0:
        return (distances[distance_valid] < thr).sum() / num_distance_valid
    return -1


def _get_max_preds(heatmaps):
    """Get keypoint predictions from score maps.

    Note:
        batch_size: N
        num_keypoints: K
        heatmap height: H
        heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.

    Returns:
        tuple: A tuple containing aggregated results.

        - preds (np.ndarray[N, K, 2]): Predicted keypoint location.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    assert isinstance(heatmaps,
                      np.ndarray), ('heatmaps should be numpy.ndarray')
    assert heatmaps.ndim == 4, 'batch_images should be 4-ndim'

    N, K, _, W = heatmaps.shape
    heatmaps_reshaped = heatmaps.reshape((N, K, -1))
    idx = np.argmax(heatmaps_reshaped, 2).reshape((N, K, 1))
    maxvals = np.amax(heatmaps_reshaped, 2).reshape((N, K, 1))

    preds = np.tile(idx, (1, 1, 2)).astype(np.float32)
    preds[:, :, 0] = preds[:, :, 0] % W
    preds[:, :, 1] = preds[:, :, 1] // W

    preds = np.where(np.tile(maxvals, (1, 1, 2)) > 0.0, preds, -1)
    return preds, maxvals


def _get_max_preds_3d(heatmaps):
    """Get keypoint predictions from 3D score maps.

    Note:
        batch size: N
        num keypoints: K
        heatmap depth size: D
        heatmap height: H
        heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, D, H, W]): model predicted heatmaps.

    Returns:
        tuple: A tuple containing aggregated results.

        - preds (np.ndarray[N, K, 3]): Predicted keypoint location.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    assert isinstance(heatmaps, np.ndarray), \
        ('heatmaps should be numpy.ndarray')
    assert heatmaps.ndim == 5, 'heatmaps should be 5-ndim'

    N, K, D, H, W = heatmaps.shape
    heatmaps_reshaped = heatmaps.reshape((N, K, -1))
    idx = np.argmax(heatmaps_reshaped, 2).reshape((N, K, 1))
    maxvals = np.amax(heatmaps_reshaped, 2).reshape((N, K, 1))

    preds = np.zeros((N, K, 3), dtype=np.float32)
    _idx = idx[..., 0]
    preds[..., 2] = _idx // (H * W)
    preds[..., 1] = (_idx // W) % H
    preds[..., 0] = _idx % W

    preds = np.where(maxvals > 0.0, preds, -1)
    return preds, maxvals


def pose_pck_accuracy(output, target, mask, thr=0.05, normalize=None):
    """Calculate the pose accuracy of PCK for each individual keypoint and the
    averaged accuracy across all keypoints from heatmaps.

    Note:
        PCK metric measures accuracy of the localization of the body joints.
        The distances between predicted positions and the ground-truth ones
        are typically normalized by the bounding box size.
        The threshold (thr) of the normalized distance is commonly set
        as 0.05, 0.1 or 0.2 etc.

        - batch_size: N
        - num_keypoints: K
        - heatmap height: H
        - heatmap width: W

    Args:
        output (np.ndarray[N, K, H, W]): Model output heatmaps.
        target (np.ndarray[N, K, H, W]): Groundtruth heatmaps.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.
        thr (float): Threshold of PCK calculation. Default 0.05.
        normalize (np.ndarray[N, 2]): Normalization factor for H&W.

    Returns:
        tuple: A tuple containing keypoint accuracy.

        - np.ndarray[K]: Accuracy of each keypoint.
        - float: Averaged accuracy across all keypoints.
        - int: Number of valid keypoints.
    """
    N, K, H, W = output.shape
    if K == 0:
        return None, 0, 0
    if normalize is None:
        normalize = np.tile(np.array([[H, W]]), (N, 1))

    pred, _ = _get_max_preds(output)
    gt, _ = _get_max_preds(target)
    return keypoint_pck_accuracy(pred, gt, mask, thr, normalize)


def keypoint_pck_accuracy(pred, gt, mask, thr, normalize):
    """Calculate the pose accuracy of PCK for each individual keypoint and the
    averaged accuracy across all keypoints for coordinates.

    Note:
        PCK metric measures accuracy of the localization of the body joints.
        The distances between predicted positions and the ground-truth ones
        are typically normalized by the bounding box size.
        The threshold (thr) of the normalized distance is commonly set
        as 0.05, 0.1 or 0.2 etc.

        - batch_size: N
        - num_keypoints: K

    Args:
        pred (np.ndarray[N, K, 2]): Predicted keypoint location.
        gt (np.ndarray[N, K, 2]): Groundtruth keypoint location.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.
        thr (float): Threshold of PCK calculation.
        normalize (np.ndarray[N, 2]): Normalization factor for H&W.

    Returns:
        tuple: A tuple containing keypoint accuracy.

        - acc (np.ndarray[K]): Accuracy of each keypoint.
        - avg_acc (float): Averaged accuracy across all keypoints.
        - cnt (int): Number of valid keypoints.
    """
    distances = _calc_distances(pred, gt, mask, normalize)

    acc = np.array([_distance_acc(d, thr) for d in distances])
    valid_acc = acc[acc >= 0]
    cnt = len(valid_acc)
    avg_acc = valid_acc.mean() if cnt > 0 else 0
    return acc, avg_acc, cnt


def keypoint_auc(pred, gt, mask, normalize, num_step=20):
    """Calculate the pose accuracy of PCK for each individual keypoint and the
    averaged accuracy across all keypoints for coordinates.

    Note:
        - batch_size: N
        - num_keypoints: K

    Args:
        pred (np.ndarray[N, K, 2]): Predicted keypoint location.
        gt (np.ndarray[N, K, 2]): Groundtruth keypoint location.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.
        normalize (float): Normalization factor.

    Returns:
        float: Area under curve.
    """
    nor = np.tile(np.array([[normalize, normalize]]), (pred.shape[0], 1))
    x = [1.0 * i / num_step for i in range(num_step)]
    y = []
    for thr in x:
        _, avg_acc, _ = keypoint_pck_accuracy(pred, gt, mask, thr, nor)
        y.append(avg_acc)

    auc = 0
    for i in range(num_step):
        auc += 1.0 / num_step * y[i]
    return auc


def keypoint_nme(pred, gt, mask, normalize_factor):
    """Calculate the normalized mean error (NME).

    Note:
        - batch_size: N
        - num_keypoints: K

    Args:
        pred (np.ndarray[N, K, 2]): Predicted keypoint location.
        gt (np.ndarray[N, K, 2]): Groundtruth keypoint location.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.
        normalize_factor (np.ndarray[N, 2]): Normalization factor.

    Returns:
        float: normalized mean error
    """
    distances = _calc_distances(pred, gt, mask, normalize_factor)
    distance_valid = distances[distances != -1]
    return distance_valid.sum() / max(1, len(distance_valid))


def keypoint_epe(pred, gt, mask):
    """Calculate the end-point error.

    Note:
        - batch_size: N
        - num_keypoints: K

    Args:
        pred (np.ndarray[N, K, 2]): Predicted keypoint location.
        gt (np.ndarray[N, K, 2]): Groundtruth keypoint location.
        mask (np.ndarray[N, K]): Visibility of the target. False for invisible
            joints, and True for visible. Invisible joints will be ignored for
            accuracy calculation.

    Returns:
        float: Average end-point error.
    """

    distances = _calc_distances(
        pred, gt, mask,
        np.ones((pred.shape[0], pred.shape[2]), dtype=np.float32))
    distance_valid = distances[distances != -1]
    return distance_valid.sum() / max(1, len(distance_valid))


def _taylor(heatmap, coord):
    """Distribution aware coordinate decoding method.

    Note:
        - heatmap height: H
        - heatmap width: W

    Args:
        heatmap (np.ndarray[H, W]): Heatmap of a particular joint type.
        coord (np.ndarray[2,]): Coordinates of the predicted keypoints.

    Returns:
        np.ndarray[2,]: Updated coordinates.
    """
    H, W = heatmap.shape[:2]
    px, py = int(coord[0]), int(coord[1])
    if 1 < px < W - 2 and 1 < py < H - 2:
        dx = 0.5 * (heatmap[py][px + 1] - heatmap[py][px - 1])
        dy = 0.5 * (heatmap[py + 1][px] - heatmap[py - 1][px])
        dxx = 0.25 * (
            heatmap[py][px + 2] - 2 * heatmap[py][px] + heatmap[py][px - 2])
        dxy = 0.25 * (
            heatmap[py + 1][px + 1] - heatmap[py - 1][px + 1] -
            heatmap[py + 1][px - 1] + heatmap[py - 1][px - 1])
        dyy = 0.25 * (
            heatmap[py + 2 * 1][px] - 2 * heatmap[py][px] +
            heatmap[py - 2 * 1][px])
        derivative = np.array([[dx], [dy]])
        hessian = np.array([[dxx, dxy], [dxy, dyy]])
        if dxx * dyy - dxy**2 != 0:
            hessianinv = np.linalg.inv(hessian)
            offset = -hessianinv @ derivative
            offset = np.squeeze(np.array(offset.T), axis=0)
            coord += offset
    return coord


def post_dark_udp(coords, batch_heatmaps, kernel=3):
    """DARK post-pocessing. Implemented by udp. Paper ref: Huang et al. The
    Devil is in the Details: Delving into Unbiased Data Processing for Human
    Pose Estimation (CVPR 2020). Zhang et al. Distribution-Aware Coordinate
    Representation for Human Pose Estimation (CVPR 2020).

    Note:
        - batch size: B
        - num keypoints: K
        - num persons: N
        - height of heatmaps: H
        - width of heatmaps: W

        B=1 for bottom_up paradigm where all persons share the same heatmap.
        B=N for top_down paradigm where each person has its own heatmaps.

    Args:
        coords (np.ndarray[N, K, 2]): Initial coordinates of human pose.
        batch_heatmaps (np.ndarray[B, K, H, W]): batch_heatmaps
        kernel (int): Gaussian kernel size (K) for modulation.

    Returns:
        np.ndarray([N, K, 2]): Refined coordinates.
    """
    if not isinstance(batch_heatmaps, np.ndarray):
        batch_heatmaps = batch_heatmaps.cpu().numpy()
    B, K, H, W = batch_heatmaps.shape
    N = coords.shape[0]
    assert (B == 1 or B == N)
    for heatmaps in batch_heatmaps:
        for heatmap in heatmaps:
            cv2.GaussianBlur(heatmap, (kernel, kernel), 0, heatmap)
    np.clip(batch_heatmaps, 0.001, 50, batch_heatmaps)
    np.log(batch_heatmaps, batch_heatmaps)

    batch_heatmaps_pad = np.pad(
        batch_heatmaps, ((0, 0), (0, 0), (1, 1), (1, 1)),
        mode='edge').flatten()

    index = coords[..., 0] + 1 + (coords[..., 1] + 1) * (W + 2)
    index += (W + 2) * (H + 2) * np.arange(0, B * K).reshape(-1, K)
    index = index.astype(int).reshape(-1, 1)
    i_ = batch_heatmaps_pad[index]
    ix1 = batch_heatmaps_pad[index + 1]
    iy1 = batch_heatmaps_pad[index + W + 2]
    ix1y1 = batch_heatmaps_pad[index + W + 3]
    ix1_y1_ = batch_heatmaps_pad[index - W - 3]
    ix1_ = batch_heatmaps_pad[index - 1]
    iy1_ = batch_heatmaps_pad[index - 2 - W]

    dx = 0.5 * (ix1 - ix1_)
    dy = 0.5 * (iy1 - iy1_)
    derivative = np.concatenate([dx, dy], axis=1)
    derivative = derivative.reshape(N, K, 2, 1)
    dxx = ix1 - 2 * i_ + ix1_
    dyy = iy1 - 2 * i_ + iy1_
    dxy = 0.5 * (ix1y1 - ix1 - iy1 + i_ + i_ - ix1_ - iy1_ + ix1_y1_)
    hessian = np.concatenate([dxx, dxy, dxy, dyy], axis=1)
    hessian = hessian.reshape(N, K, 2, 2)
    hessian = np.linalg.inv(hessian + np.finfo(np.float32).eps * np.eye(2))
    coords -= np.einsum('ijmn,ijnk->ijmk', hessian, derivative).squeeze()
    return coords


def _gaussian_blur(heatmaps, kernel=11):
    """Modulate heatmap distribution with Gaussian.
     sigma = 0.3*((kernel_size-1)*0.5-1)+0.8
     sigma~=3 if k=17
     sigma=2 if k=11;
     sigma~=1.5 if k=7;
     sigma~=1 if k=3;

    Note:
        - batch_size: N
        - num_keypoints: K
        - heatmap height: H
        - heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.
        kernel (int): Gaussian kernel size (K) for modulation, which should
            match the heatmap gaussian sigma when training.
            K=17 for sigma=3 and k=11 for sigma=2.

    Returns:
        np.ndarray ([N, K, H, W]): Modulated heatmap distribution.
    """
    assert kernel % 2 == 1

    border = (kernel - 1) // 2
    batch_size = heatmaps.shape[0]
    num_joints = heatmaps.shape[1]
    height = heatmaps.shape[2]
    width = heatmaps.shape[3]
    for i in range(batch_size):
        for j in range(num_joints):
            origin_max = np.max(heatmaps[i, j])
            dr = np.zeros((height + 2 * border, width + 2 * border),
                          dtype=np.float32)
            dr[border:-border, border:-border] = heatmaps[i, j].copy()
            dr = cv2.GaussianBlur(dr, (kernel, kernel), 0)
            heatmaps[i, j] = dr[border:-border, border:-border].copy()
            heatmaps[i, j] *= origin_max / np.max(heatmaps[i, j])
    return heatmaps


def keypoints_from_regression(regression_preds, center, scale, img_size):
    """Get final keypoint predictions from regression vectors and transform
    them back to the image.

    Note:
        - batch_size: N
        - num_keypoints: K

    Args:
        regression_preds (np.ndarray[N, K, 2]): model prediction.
        center (np.ndarray[N, 2]): Center of the bounding box (x, y).
        scale (np.ndarray[N, 2]): Scale of the bounding box
            wrt height/width.
        img_size (list(img_width, img_height)): model input image size.

    Returns:
        tuple:

        - preds (np.ndarray[N, K, 2]): Predicted keypoint location in images.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    N, K, _ = regression_preds.shape
    preds, maxvals = regression_preds, np.ones((N, K, 1), dtype=np.float32)

    preds = preds * img_size

    # Transform back to the image
    for i in range(N):
        preds[i] = transform_preds(preds[i], center[i], scale[i], img_size)

    return preds, maxvals


def keypoints_from_heatmaps(heatmaps,
                            center,
                            scale,
                            unbiased=False,
                            post_process='default',
                            kernel=11,
                            valid_radius_factor=0.0546875,
                            use_udp=False,
                            target_type='GaussianHeatmap'):
    """Get final keypoint predictions from heatmaps and transform them back to
    the image.

    Note:
        - batch size: N
        - num keypoints: K
        - heatmap height: H
        - heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.
        center (np.ndarray[N, 2]): Center of the bounding box (x, y).
        scale (np.ndarray[N, 2]): Scale of the bounding box
            wrt height/width.
        post_process (str/None): Choice of methods to post-process
            heatmaps. Currently supported: None, 'default', 'unbiased',
            'megvii'.
        unbiased (bool): Option to use unbiased decoding. Mutually
            exclusive with megvii.
            Note: this arg is deprecated and unbiased=True can be replaced
            by post_process='unbiased'
            Paper ref: Zhang et al. Distribution-Aware Coordinate
            Representation for Human Pose Estimation (CVPR 2020).
        kernel (int): Gaussian kernel size (K) for modulation, which should
            match the heatmap gaussian sigma when training.
            K=17 for sigma=3 and k=11 for sigma=2.
        valid_radius_factor (float): The radius factor of the positive area
            in classification heatmap for UDP.
        use_udp (bool): Use unbiased data processing.
        target_type (str): 'GaussianHeatmap' or 'CombinedTarget'.
            GaussianHeatmap: Classification target with gaussian distribution.
            CombinedTarget: The combination of classification target
            (response map) and regression target (offset map).
            Paper ref: Huang et al. The Devil is in the Details: Delving into
            Unbiased Data Processing for Human Pose Estimation (CVPR 2020).

    Returns:
        tuple: A tuple containing keypoint predictions and scores.

        - preds (np.ndarray[N, K, 2]): Predicted keypoint location in images.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    # Avoid being affected
    heatmaps = heatmaps.copy()

    # detect conflicts
    if unbiased:
        assert post_process not in [False, None, 'megvii']
    if post_process in ['megvii', 'unbiased']:
        assert kernel > 0
    if use_udp:
        assert not post_process == 'megvii'

    # normalize configs
    if post_process is False:
        warnings.warn(
            'post_process=False is deprecated, '
            'please use post_process=None instead', DeprecationWarning)
        post_process = None
    elif post_process is True:
        if unbiased is True:
            warnings.warn(
                'post_process=True, unbiased=True is deprecated,'
                " please use post_process='unbiased' instead",
                DeprecationWarning)
            post_process = 'unbiased'
        else:
            warnings.warn(
                'post_process=True, unbiased=False is deprecated, '
                "please use post_process='default' instead",
                DeprecationWarning)
            post_process = 'default'
    elif post_process == 'default':
        if unbiased is True:
            warnings.warn(
                'unbiased=True is deprecated, please use '
                "post_process='unbiased' instead", DeprecationWarning)
            post_process = 'unbiased'

    # start processing
    if post_process == 'megvii':
        heatmaps = _gaussian_blur(heatmaps, kernel=kernel)

    N, K, H, W = heatmaps.shape
    if use_udp:
        if target_type.lower() == 'GaussianHeatMap'.lower():
            preds, maxvals = _get_max_preds(heatmaps)
            preds = post_dark_udp(preds, heatmaps, kernel=kernel)
        elif target_type.lower() == 'CombinedTarget'.lower():
            for person_heatmaps in heatmaps:
                for i, heatmap in enumerate(person_heatmaps):
                    kt = 2 * kernel + 1 if i % 3 == 0 else kernel
                    cv2.GaussianBlur(heatmap, (kt, kt), 0, heatmap)
            # valid radius is in direct proportion to the height of heatmap.
            valid_radius = valid_radius_factor * H
            offset_x = heatmaps[:, 1::3, :].flatten() * valid_radius
            offset_y = heatmaps[:, 2::3, :].flatten() * valid_radius
            heatmaps = heatmaps[:, ::3, :]
            preds, maxvals = _get_max_preds(heatmaps)
            index = preds[..., 0] + preds[..., 1] * W
            index += W * H * np.arange(0, N * K / 3)
            index = index.astype(int).reshape(N, K // 3, 1)
            preds += np.concatenate((offset_x[index], offset_y[index]), axis=2)
        else:
            raise ValueError('target_type should be either '
                             "'GaussianHeatmap' or 'CombinedTarget'")
    else:
        preds, maxvals = _get_max_preds(heatmaps)
        if post_process == 'unbiased':  # alleviate biased coordinate
            # apply Gaussian distribution modulation.
            heatmaps = np.log(
                np.maximum(_gaussian_blur(heatmaps, kernel), 1e-10))
            for n in range(N):
                for k in range(K):
                    preds[n][k] = _taylor(heatmaps[n][k], preds[n][k])
        elif post_process is not None:
            # add +/-0.25 shift to the predicted locations for higher acc.
            for n in range(N):
                for k in range(K):
                    heatmap = heatmaps[n][k]
                    px = int(preds[n][k][0])
                    py = int(preds[n][k][1])
                    if 1 < px < W - 1 and 1 < py < H - 1:
                        diff = np.array([
                            heatmap[py][px + 1] - heatmap[py][px - 1],
                            heatmap[py + 1][px] - heatmap[py - 1][px]
                        ])
                        preds[n][k] += np.sign(diff) * .25
                        if post_process == 'megvii':
                            preds[n][k] += 0.5

    # Transform back to the image
    for i in range(N):
        preds[i] = transform_preds(
            preds[i], center[i], scale[i], [W, H], use_udp=use_udp)

    if post_process == 'megvii':
        maxvals = maxvals / 255.0 + 0.5

    return preds, maxvals


def keypoints_from_heatmaps3d(heatmaps, center, scale):
    """Get final keypoint predictions from 3d heatmaps and transform them back
    to the image.

    Note:
        - batch size: N
        - num keypoints: K
        - heatmap depth size: D
        - heatmap height: H
        - heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, D, H, W]): model predicted heatmaps.
        center (np.ndarray[N, 2]): Center of the bounding box (x, y).
        scale (np.ndarray[N, 2]): Scale of the bounding box
            wrt height/width.

    Returns:
        tuple: A tuple containing keypoint predictions and scores.

        - preds (np.ndarray[N, K, 3]): Predicted 3d keypoint location \
            in images.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    N, K, D, H, W = heatmaps.shape
    preds, maxvals = _get_max_preds_3d(heatmaps)
    # Transform back to the image
    for i in range(N):
        preds[i, :, :2] = transform_preds(preds[i, :, :2], center[i], scale[i],
                                          [W, H])
    return preds, maxvals


def multilabel_classification_accuracy(pred, gt, mask, thr=0.5):
    """Get multi-label classification accuracy.

    Note:
        - batch size: N
        - label number: L

    Args:
        pred (np.ndarray[N, L, 2]): model predicted labels.
        gt (np.ndarray[N, L, 2]): ground-truth labels.
        mask (np.ndarray[N, 1] or np.ndarray[N, L] ): reliability of
        ground-truth labels.

    Returns:
        float: multi-label classification accuracy.
    """
    # we only compute accuracy on the samples with ground-truth of all labels.
    valid = (mask > 0).min(axis=1) if mask.ndim == 2 else (mask > 0)
    pred, gt = pred[valid], gt[valid]

    if pred.shape[0] == 0:
        acc = 0.0  # when no sample is with gt labels, set acc to 0.
    else:
        # The classification of a sample is regarded as correct
        # only if it's correct for all labels.
        acc = (((pred - thr) * (gt - thr)) > 0).all(axis=1).mean()
    return acc



import torch.nn as nn

from .backbone.vit import ViT
from .head.topdown_heatmap_simple_head import TopdownHeatmapSimpleHead


__all__ = ['ViTPose']


class ViTPose(nn.Module):
    def __init__(self, cfg: dict) -> None:
        super(ViTPose, self).__init__()
        
        backbone_cfg = {k: v for k, v in cfg['backbone'].items() if k != 'type'}
        head_cfg = {k: v for k, v in cfg['keypoint_head'].items() if k != 'type'}
        
        self.backbone = ViT(**backbone_cfg)
        self.keypoint_head = TopdownHeatmapSimpleHead(**head_cfg)
    
    def forward_features(self, x):
        return self.backbone(x)
    
    def forward(self, x):
        return self.keypoint_head(self.backbone(x))


from collections import defaultdict
from filterpy.kalman import KalmanFilter
from catboost import CatBoostRegressor
import pandas as pd
import cv2
import numpy as np
import os
from ..constants import WEIGHTS_COCO


class KeypointTracker:
    def __init__(self, skeletons_init, var_smooth=(100, 100), var_pred=(100, 20), wait_recovery=10,
                 keypoint_weights=WEIGHTS_COCO):
        self.ids = defaultdict()
        self.var_smooth = var_smooth
        self.var_pred = var_pred
        self.frame = 0
        self.wait_recovery = wait_recovery
        self.model = CatBoostRegressor()
        self.model.load_model(os.path.join(os.path.dirname(__file__), 'reg_model_1'))
        self.cur_max_id = -1
        self.keypoint_weights = keypoint_weights

        for sk in skeletons_init:
            self.ids[self.cur_max_id] = self.create_id(sk)

    def step(self, skeletons):
        self.frame += 1
        self.match, self.map_ids, self.free_ids, self.free_new_ids = self.matching(skeletons)
        for id_, index in self.map_ids:
            new_sk = skeletons[index]
            self.ids[id_]['frame'] = self.frame
            self.ids[id_]['skeleton'] = new_sk
            prediction = self.ids[id_]['predictor'].predict(new_sk[:, :2])
            self.ids[id_]['prediction'] = prediction

        for id_ in self.free_ids:
            # id recovering
            if self.ids[id_]['prediction'] is not None:
                pred_sk = self.ids[id_]['skeleton'][:, :2] + self.ids[id_]['prediction']
                self.ids[id_]['skeleton'] = pred_sk
                prediction = self.ids[id_]['predictor'].predict(pred_sk)
                self.ids[id_]['prediction'] = prediction
            else:
                del self.ids[id_]

        for new_id in self.free_new_ids:
            sk = skeletons[new_id]
            self.ids[self.cur_max_id] = self.create_id(sk)

        self.index_to_del = [x for x in self.ids if self.frame - self.ids[x]['frame'] > self.wait_recovery]
        for index in self.index_to_del:
            del self.ids[index]

    def create_id(self, sk, prediction=None):
        self.cur_max_id += 1
        return defaultdict(None, [('id', self.cur_max_id), ('frame', self.frame), ('skeleton', sk),
                                  ('predictor', KF_skeleton_predictor(sk[:, :2], self.var_smooth, self.var_pred)),
                                  ('prediction', prediction)])

    def current(self):
        ids = [id_ for id_ in self.ids if self.ids[id_]['frame'] == self.frame]
        skeletons = np.float64([self.ids[id_]['skeleton'] for id_ in ids])
        return skeletons, ids

    def matching(self, skeletons, tresh=0.5):
        match = pd.DataFrame()
        for id_ in self.ids:
            skeleton1 = self.ids[id_]['skeleton']
            prediction = self.ids[id_]['prediction']
            for j, skeleton2 in enumerate(skeletons):
                X, ret = self.get_statistics(skeleton1, skeleton2, prediction)
                if ret:
                    if prediction is None:
                        match.loc[id_, j] = X[-1] ** (-1)
                    else:
                        match.loc[id_, j] = self.model.predict(X)

        match = match[match > tresh].sort_index(axis=1)
        map_ids, free_ids, free_new_ids = self.get_indexes_from(match)
        return match, map_ids, free_ids, free_new_ids

    def get_statistics(self, skeleton1, skeleton2, prediction=None, ratio=3.6, amount=0.1):
        skeleton1, skeleton2 = skeleton1[:, :2], skeleton2[:, :2]  # TODO: append confidence dependency

        ft = fluctuation_threshold(skeleton1)
        d_sk = skeleton_distance(skeleton1, skeleton2)
        r, phi = xy2rphi(V(skeleton2 - skeleton1, self.keypoint_weights))
        if prediction is None:
            return [1, np.inf, 0, 1, np.inf, 0, d_sk / ft], locality_criterion(r, ft, ratio, amount)
        else:
            d_sk_pred = skeleton_distance(skeleton1 + prediction, skeleton2)
            d_CM_pred = CM_distance(skeleton1 + prediction, skeleton2, self.keypoint_weights)
            r_pred, phi_pred = xy2rphi(V(prediction, self.keypoint_weights))

            angle = np.abs(phi - phi_pred)
            angle = 360 - angle if angle > 180 else angle

        return [d_sk_pred / d_sk, r / r_pred, angle, d_CM_pred / r, d_CM_pred / r_pred, r_pred / ft,
                d_sk / ft], locality_criterion(r, ft, ratio, amount)

    @staticmethod
    def get_indexes_from(match):
        idx = np.array(list(match.stack().sort_values(ascending=False).index))

        axis1, axis2 = [], []
        for a, b in idx:
            if (a not in axis1) and (b not in axis2):
                axis1.append(a), axis2.append(b)

        free_ids = [i for i in match.index if i not in axis1]
        free_new_ids = [i for i in match.columns if i not in axis2]
        map_ids = pd.DataFrame([axis1, axis2]).T.sort_values(0).to_numpy()
        return map_ids, free_ids, free_new_ids


def locality_criterion(r, ft, ratio=3.6, amount=0.1):
    return r < ratio * ft / amount


class KF:
    def __init__(self, pos=0.0, var1=50, var2=10):
        self.f = KalmanFilter(dim_x=2, dim_z=1)
        self.f.x = np.array([pos, 0.])
        self.f.F = np.array([[1., 1.],
                             [0., 1.]])
        self.f.H = np.array([[1., 0.]])
        self.f.P = np.array([[var1, 0.],
                             [0., var1]])
        self.f.R = np.array([[var2]])

    def predict(self):
        self.f.predict()

    def update(self, z):
        self.f.update(z)

    def get_value(self):
        return self.f.x[0]


class KF_smooth_predictor:
    def __init__(self, x=0, var_smooth=(1000, 1000), var_pred=(100, 20)):
        self.kf_smooth = KF(x, *var_smooth)
        self.prev = x
        self.state = 0
        self.var_pred = var_pred

    def diff_smooth(self, x):

        self.kf_smooth.predict()
        self.kf_smooth.update(x)
        diff = self.kf_smooth.get_value() - self.prev
        self.prev = self.kf_smooth.get_value()

        return diff

    def predict(self, x=None):
        if self.state == 0:

            self.diff = self.diff_smooth(x)
            self.kf_pred = KF(self.diff, *self.var_pred)
            self.state = 1

        else:
            self.kf_pred.update(self.diff_smooth(x))
            self.kf_pred.predict()

        return self.kf_pred.get_value()


class KF_smooth_predictor_2D:
    def __init__(self, x=0, y=0, var_smooth=(1000, 1000), var_pred=(100, 20)):
        self.kfx = KF_smooth_predictor(x, var_smooth, var_pred)
        self.kfy = KF_smooth_predictor(y, var_smooth, var_pred)

    def predict(self, x, y):
        return [self.kfx.predict(x), self.kfy.predict(y)]


class KF_skeleton_predictor:
    def __init__(self, skeleton, var_smooth=(100, 100), var_pred=(100, 20)):
        self.skeleton_predictor = defaultdict()
        self.shape = skeleton.shape

        for i, point in enumerate(skeleton):
            self.skeleton_predictor[i] = KF_smooth_predictor_2D(*point, var_smooth, var_pred)

    def predict(self, skeleton):
        assert skeleton.shape == self.shape

        preds = []
        for i, point in enumerate(skeleton):
            preds.append(self.skeleton_predictor[i].predict(*point))

        return np.array(preds)


def xy2rphi(points):
    points = np.array(points)
    x, y = points.T
    r = np.sqrt(x ** 2 + y ** 2)
    phi = np.degrees(np.arctan2(y, x))
    return np.array([r, phi]).T


def ceneter_mass(skeleton, keypoint_weights):
    return skeleton.T @ keypoint_weights


def V(prediction, keypoint_weights):
    return prediction.T @ keypoint_weights


def skeleton_distance(skeleton1, skeleton2):
    return np.sqrt(((skeleton1 - skeleton2) ** 2).sum(axis=1)).mean()


def CM_distance(skeleton1, skeleton2, keypoint_weights):
    return np.sqrt(
        ((ceneter_mass(skeleton1, keypoint_weights) - ceneter_mass(skeleton2, keypoint_weights)) ** 2).sum(axis=0))


def fluctuation_threshold(skeleton, amount=0.1):
    return (skeleton.max(0) - skeleton.min(0)).mean() * amount


def crop_person(skeleton, img, pad=20):
    x0, y0, x1, y1 = np.int32([skeleton.min(0), skeleton.max(0)]).flatten()
    return img[y0 - pad:y1 + pad, x0 - pad:x1 + pad]


def crop_person_fixed(skeleton, img, keypoint_weights, crop=100):
    x, y = ceneter_mass(skeleton, keypoint_weights).astype(int)
    return img[y - crop:y + crop, x - crop:x + crop]


def show_pred(skeleton, pred, img):
    for x, y, dx, dy in np.hstack((skeleton, pred)):
        cv2.line(img, (int(x), int(y)), (int(x + dx), int(y + dy)), (255, 0, 0), thickness=1)
        cv2.circle(img, (int(x + dx), int(y + dy)), thickness=-1, color=(255, 0, 0), radius=1)
    return img


def show_V(skeleton, pred, img, keypoint_weights):
    x, y = ceneter_mass(skeleton, keypoint_weights).astype(int)
    dx, dy = V(pred)
    cv2.line(img, (int(x), int(y)), (int(x + dx), int(y + dy)), (255, 0, 0), thickness=1)
    cv2.circle(img, (int(x + dx), int(y + dy)), thickness=-1, color=(255, 0, 0), radius=1)
    return img



import numpy as np
import torch


def normalize_screen_coordinates(X, w, h):
    assert X.shape[-1] == 2
    return X / w * 2 - [1, h / w]


def world_to_camera(X, R, t):
    Rt = wrap(qinverse, R)
    return wrap(qrot, np.tile(Rt, (*X.shape[:-1], 1)), X - t)


def camera_to_world(X, R, t):
    return wrap(qrot, np.tile(R, (*X.shape[:-1], 1)), X) + t


def wrap(func, *args, unsqueeze=False):
    args = list(args)
    for i, arg in enumerate(args):
        if type(arg) == np.ndarray:
            args[i] = torch.from_numpy(arg)
            if unsqueeze:
                args[i] = args[i].unsqueeze(0)

    result = func(*args)

    if isinstance(result, tuple):
        result = list(result)
        for i, res in enumerate(result):
            if type(res) == torch.Tensor:
                if unsqueeze:
                    res = res.squeeze(0)
                result[i] = res.numpy()
        return tuple(result)
    elif type(result) == torch.Tensor:
        if unsqueeze:
            result = result.squeeze(0)
        return result.numpy()
    else:
        return result


def qrot(q, v):
    assert q.shape[-1] == 4
    assert v.shape[-1] == 3
    assert q.shape[:-1] == v.shape[:-1]

    qvec = q[..., 1:]
    uv = torch.cross(qvec, v, dim=len(q.shape) - 1)
    uuv = torch.cross(qvec, uv, dim=len(q.shape) - 1)
    return (v + 2 * (q[..., :1] * uv + uuv))


def qinverse(q, inplace=False):
    if inplace:
        q[..., 1:] *= -1
        return q
    else:
        w = q[..., :1]
        xyz = q[..., 1:]
        return torch.cat((w, -xyz), dim=len(q.shape) - 1)



import math
import warnings

from itertools import repeat
import collections.abc

import torch
from functools import partial
import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from torch import Tensor

# from timm.models.layers import drop_path, to_2tuple, trunc_normal_

# from .base_backbone import BaseBackbone

def drop_path(x, drop_prob: float = 0., training: bool = False, scale_by_keep: bool = True):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor

def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
            return x
        return tuple(repeat(x, n))
    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple

def _trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the normal distribution.
    # Get upper and lower cdf values
    l = norm_cdf((a - mean) / std)
    u = norm_cdf((b - mean) / std)

    # Uniformly fill tensor with values from [l, u], then translate to
    # [2l-1, 2u-1].
    tensor.uniform_(2 * l - 1, 2 * u - 1)

    # Use inverse cdf transform for normal distribution to get truncated
    # standard normal
    tensor.erfinv_()

    # Transform to proper mean, std
    tensor.mul_(std * math.sqrt(2.))
    tensor.add_(mean)

    # Clamp to ensure it's in the proper range
    tensor.clamp_(min=a, max=b)
    return tensor


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    # type: (Tensor, float, float, float, float) -> Tensor
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.

    NOTE: this impl is similar to the PyTorch trunc_normal_, the bounds [a, b] are
    applied while sampling the normal with mean/std applied, therefore a, b args
    should be adjusted to match the range of mean, std args.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    with torch.no_grad():
        return _trunc_normal_(tensor, mean, std, a, b)

class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)
    
    def extra_repr(self):
        return 'p={}'.format(self.drop_prob)

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class Attention(nn.Module):
    def __init__(
            self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.,
            proj_drop=0., attn_head_dim=None,):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.dim = dim

        if attn_head_dim is not None:
            head_dim = attn_head_dim
        all_head_dim = head_dim * self.num_heads

        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, all_head_dim * 3, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(all_head_dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x)
        qkv = qkv.reshape(B, N, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x

class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, 
                 drop=0., attn_drop=0., drop_path=0., act_layer=nn.GELU, 
                 norm_layer=nn.LayerNorm, attn_head_dim=None
                 ):
        super().__init__()
        
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, attn_head_dim=attn_head_dim
            )

        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768, ratio=1):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0]) * (ratio ** 2)
        self.patch_shape = (int(img_size[0] // patch_size[0] * ratio), int(img_size[1] // patch_size[1] * ratio))
        self.origin_patch_shape = (int(img_size[0] // patch_size[0]), int(img_size[1] // patch_size[1]))
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=(patch_size[0] // ratio), padding=4 + 2 * (ratio//2-1))

    def forward(self, x, **kwargs):
        B, C, H, W = x.shape
        x = self.proj(x)
        Hp, Wp = x.shape[2], x.shape[3]

        x = x.flatten(2).transpose(1, 2)
        return x, (Hp, Wp)


class HybridEmbed(nn.Module):
    """ CNN Feature Map Embedding
    Extract feature map from CNN, flatten, project to embedding dim.
    """
    def __init__(self, backbone, img_size=224, feature_size=None, in_chans=3, embed_dim=768):
        super().__init__()
        assert isinstance(backbone, nn.Module)
        img_size = to_2tuple(img_size)
        self.img_size = img_size
        self.backbone = backbone
        if feature_size is None:
            with torch.no_grad():
                training = backbone.training
                if training:
                    backbone.eval()
                o = self.backbone(torch.zeros(1, in_chans, img_size[0], img_size[1]))[-1]
                feature_size = o.shape[-2:]
                feature_dim = o.shape[1]
                backbone.train(training)
        else:
            feature_size = to_2tuple(feature_size)
            feature_dim = self.backbone.feature_info.channels()[-1]
        self.num_patches = feature_size[0] * feature_size[1]
        self.proj = nn.Linear(feature_dim, embed_dim)

    def forward(self, x):
        x = self.backbone(x)[-1]
        x = x.flatten(2).transpose(1, 2)
        x = self.proj(x)
        return x


class ViT(nn.Module):
    def __init__(self,
                 img_size=224, patch_size=16, in_chans=3, num_classes=80, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0., hybrid_backbone=None, norm_layer=None, use_checkpoint=False, 
                 frozen_stages=-1, ratio=1, last_norm=True,
                 patch_padding='pad', freeze_attn=False, freeze_ffn=False,
                 ):
        super(ViT, self).__init__()
        # Protect mutable default arguments
        
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        self.num_classes = num_classes
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.frozen_stages = frozen_stages
        self.use_checkpoint = use_checkpoint
        self.patch_padding = patch_padding
        self.freeze_attn = freeze_attn
        self.freeze_ffn = freeze_ffn
        self.depth = depth

        if hybrid_backbone is not None:
            self.patch_embed = HybridEmbed(
                hybrid_backbone, img_size=img_size, in_chans=in_chans, embed_dim=embed_dim)
        else:
            self.patch_embed = PatchEmbed(
                img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim, ratio=ratio)
        num_patches = self.patch_embed.num_patches

        # since the pretraining model has class token
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule

        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer,
                )
            for i in range(depth)])

        self.last_norm = norm_layer(embed_dim) if last_norm else nn.Identity()

        if self.pos_embed is not None:
            trunc_normal_(self.pos_embed, std=.02)

        self._freeze_stages()

    def _freeze_stages(self):
        """Freeze parameters."""
        if self.frozen_stages >= 0:
            self.patch_embed.eval()
            for param in self.patch_embed.parameters():
                param.requires_grad = False

        for i in range(1, self.frozen_stages + 1):
            m = self.blocks[i]
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

        if self.freeze_attn:
            for i in range(0, self.depth):
                m = self.blocks[i]
                m.attn.eval()
                m.norm1.eval()
                for param in m.attn.parameters():
                    param.requires_grad = False
                for param in m.norm1.parameters():
                    param.requires_grad = False

        if self.freeze_ffn:
            self.pos_embed.requires_grad = False
            self.patch_embed.eval()
            for param in self.patch_embed.parameters():
                param.requires_grad = False
            for i in range(0, self.depth):
                m = self.blocks[i]
                m.mlp.eval()
                m.norm2.eval()
                for param in m.mlp.parameters():
                    param.requires_grad = False
                for param in m.norm2.parameters():
                    param.requires_grad = False

    def init_weights(self, pretrained=None):
        """Initialize the weights in backbone.
        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """
        super().init_weights(pretrained, patch_padding=self.patch_padding)

        if pretrained is None:
            def _init_weights(m):
                if isinstance(m, nn.Linear):
                    trunc_normal_(m.weight, std=.02)
                    if isinstance(m, nn.Linear) and m.bias is not None:
                        nn.init.constant_(m.bias, 0)
                elif isinstance(m, nn.LayerNorm):
                    nn.init.constant_(m.bias, 0)
                    nn.init.constant_(m.weight, 1.0)

            self.apply(_init_weights)

    def get_num_layers(self):
        return len(self.blocks)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token'}

    def forward_features(self, x):
        B, C, H, W = x.shape
        x, (Hp, Wp) = self.patch_embed(x)

        if self.pos_embed is not None:
            # fit for multiple GPU training
            # since the first element for pos embed (sin-cos manner) is zero, it will cause no difference
            x = x + self.pos_embed[:, 1:] + self.pos_embed[:, :1]

        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)

        x = self.last_norm(x)

        xp = x.permute(0, 2, 1).reshape(B, -1, Hp, Wp).contiguous()

        return xp

    def forward(self, x):
        x = self.forward_features(x)
        return x

    def train(self, mode=True):
        """Convert the model into training mode."""
        super().train(mode)
        self._freeze_stages()



from .Networks import YOLOv7, VITPOSE, MHFORMER
from .KeypointTracker import KeypointTracker
from .constants import WEIGHTS_COCO
from .utils import OpenVideo, VideoDataKeypoints, recover_kps, merge_track, poses2scene, clear
from .HumanVisualizer import Visualizer
import numpy as np
import torch
import os


class HITrack:
    def __init__(self, video_path, wait_recovery=20):
        self.wait_recovery = wait_recovery
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.video_path = video_path
        self.npz_path = ''.join(video_path.split('.')[:-1])+'.npz'
        self.data = VideoDataKeypoints(self.npz_path)

        if os.path.exists(self.npz_path):
            self.data.load()
            print(f'Loaded from file {self.npz_path}')

    def __call__(self, merge_dict=None):
        if self.data.keypoints_2d is None:
            self.compute_2d()
        self.recover_2d(merge_dict)
        self.compute_3d()
        self.compute_scene()
        self.visualize('3D_scene', original_sound=True)

    def compute_2d(self, yolo='yolov7', vitpose='b', save=True):
        self.ht = HumanTrack(self.video_path, yolo, vitpose, self.wait_recovery, self.device)
        keypoints_2d = self.ht.compute()
        clear()
        self.data.update(keypoints_2d, save=save)

    def recover_2d(self, merge_dict=None, save=True):
        keypoints_2d = self.data.keypoints_2d
        if merge_dict is not None:
            keypoints_2d = merge_track(keypoints_2d, merge_dict)

        keypoints_2d = recover_kps(keypoints_2d, self.wait_recovery)
        self.data.update(keypoints_2d, save=save)

    def compute_3d(self, mhformer='351', save=True):
        lifter = MHFORMER(mhformer, device=self.device)
        keypoints_3d = lifter(self.data.keypoints_2d)
        self.data.update(keypoints_3d=keypoints_3d, save=save)

    def compute_scene(self, save=True):
        scene = poses2scene(self.data)
        self.data.update(scene=scene, save=save)

    def visualize(self, how='3D_scene', id_=None, skeleton_format="H36", end=None, compress=False, original_sound=False):
        vis = Visualizer(self.data, video_path=self.video_path, skeleton_format=skeleton_format)
        vis(how, id_, end, compress, original_sound)


class HumanTrack:
    def __init__(self, video_path, yolo_model='yolov7', vitpose_model='b', wait_recovery=15, device='cuda'):

        self.video_path = video_path
        self.device = device

        self.wait_recovery = wait_recovery
        self.keypoint_weights = WEIGHTS_COCO
        
        self.det = YOLOv7(yolo_model, device=device)
        self.pose = VITPOSE(vitpose_model, device=device)
        self.keypoints = []

    def compute(self):
        cap = OpenVideo(self.video_path)
        for i in range(cap.length):
            rgb = cap.read()
            boxes = self.det(rgb)
            skeletons = self.pose(rgb, boxes)

            if i == 0:
                self.t = KeypointTracker(skeletons, wait_recovery=self.wait_recovery,
                                         keypoint_weights=self.keypoint_weights)
            else:
                self.t.step(skeletons)

            skeletons, ids = self.t.current()
            self.keypoints.append([i, skeletons, ids])

        keypoints_2d = np.zeros((self.t.cur_max_id + 1, cap.length, len(self.keypoint_weights), 2))
        for i, skeletons, ids in self.keypoints:
            keypoints_2d[ids, i] = skeletons[:, :, :2]

        del [self.det, self.pose]
        return keypoints_2d


import cv2
import numpy as np
from ..constants import PERSON_SKELETONS, COLORS

# import matplotlib
# COLORS = np.int32(255*np.array([matplotlib.cm.get_cmap('tab20')((x + 0.05) / 20)[:3] for x in range(20)])).tolist()


class HumanVisualizer:
    def __init__(self, label_scale=1.2, thickness=3, alpha=0.5, box_size=20, thickness_text=2, skeleton_format='COCO'):
        
        self.colors = COLORS
        self.label_scale = label_scale
        self.thickness = thickness
        self.thickness_text = thickness_text
        self.box_size = box_size
        self.alpha = alpha
        self.pad = label_scale*box_size
        self.point_size = round(1.5*thickness)
        self.skeleton_format = skeleton_format
        self.SKELETON = np.array(PERSON_SKELETONS[skeleton_format])-1

    def __call__(self, image, skeletons, ids=None):
        
        if ids is not None:
            assert len(skeletons) == len(ids)
            for id_, skeleton in zip(ids, skeletons):
                image = self.print_skeleton(image, skeleton, id_)
        else:
            for skeleton in skeletons:
                image = self.print_skeleton(image, skeleton)
                
        return image
    
    def print_skeleton(self, image, skeleton, id_=None):

        assert image.dtype == 'uint8'
        x, y = skeleton.T.astype(int)

        for ci, (j1i, j2i) in enumerate(self.SKELETON):
            p1, p2 = (x[j1i], y[j1i]), (x[j2i], y[j2i])
            cv2.circle(image, p1, self.point_size, (255,255,255), -1, cv2.LINE_AA)
            cv2.circle(image, p2, self.point_size, (255,255,255), -1, cv2.LINE_AA)
            cv2.line(image, p1, p2, self.colors[ci % 20], self.thickness, cv2.LINE_AA)
            
        if id_ is not None:    
            self.label(image, self.label_coord(x, y), id_)
            
        return image
        
    def label(self, img, point, id_):
        
        point = (int(point[0]+self.pad), int(point[1]-self.pad))
        (dx, dy) = cv2.getTextSize(str(id_), cv2.FONT_HERSHEY_SIMPLEX, self.label_scale, self.thickness)[0]
        box_coords = (point[0]-self.pad, point[1] - dy-self.pad,
                      point[0] + dx+self.pad, point[1]+self.pad)
        self.rectangle(img, box_coords, id_)
        cv2.putText(img, str(id_), (point[0], point[1]), cv2.FONT_HERSHEY_SIMPLEX, self.label_scale,
                (255, 255, 255), self.thickness_text, cv2.LINE_AA)
        
        return img
            
    def rectangle(self, image, box, id_):
        
        x1, y1, x2, y2 = np.clip(np.int32(box), 0, None)
        crop = image[y1:y2, x1:x2]
        rect = np.uint8(np.ones(crop.shape)*self.colors[id_ % 20])
        image[y1:y2, x1:x2] = crop*(1-self.alpha)+rect*self.alpha
        
        return image

    def label_coord(self, x, y):
        coord_x = x.mean()
        if self.skeleton_format == 'COCO':
            coord_y = y[:5].mean()
        elif self.skeleton_format == 'H36':
            coord_y = y[10].mean()
        else:
            raise NotImplementedError('Use only COCO or H36 format')
        return coord_x, coord_y


from setuptools import setup, find_packages
import os

def readme():
  with open('README.md', 'r') as f:
    text = ''.join([line for line in f if not line.startswith('http')])
    return text

def _parse_requirements(path):
  with open(path) as f:
    return [
        line.rstrip()
        for line in f
        if not (line.isspace() or line.startswith('#') or line.startswith('-'))
    ]

setup(
  name='HITrack',
  version='1.0.1',
  author='Kirill Shokhin',
  author_email='kashokhin@gmail.com',
  license='MIT',
  description='3D scene on a monocular video',
  long_description=readme(),
  long_description_content_type='text/markdown',
  url='https://github.com/Kirill-Shokhin/HITrack',
  packages=find_packages(),
  install_requires=_parse_requirements('requirements.txt'),
  classifiers=[
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent'
  ],
  keywords='HITrack',
  python_requires='>=3.7'
)


# Copyright (c) OpenMMLab. All rights reserved.
# from .nms import oks_iou, oks_nms, soft_oks_nms
# from .one_euro_filter import OneEuroFilter
from .post_transforms import (affine_transform, flip_back, fliplr_joints,
                              fliplr_regression, get_affine_transform,
                              get_warp_matrix, rotate_point, transform_preds,
                              warp_affine_joints)

# __all__ = [
#     'oks_nms', 'soft_oks_nms', 'affine_transform', 'rotate_point', 'flip_back',
#     'fliplr_joints', 'fliplr_regression', 'transform_preds',
#     'get_affine_transform', 'get_warp_matrix', 'warp_affine_joints',
#     'OneEuroFilter', 'oks_iou'
# ]



from .MHFormer.mhformer import Model
from .MHFormer.camera import camera_to_world
from HITrack.utils import coco2h36, clear, download_models

from torch.utils.data import DataLoader
from tqdm.notebook import tqdm
import torch
import numpy as np
import os

mhformer_model_map = {
    '351': ('model_4294.pth', '1c-STVvF_C-NdZXWjH-ZbXpWsN95fcMCU'),
    '243': ('model_4317.pth', '1xAl_lG_tFBeyN6R8OYh2apdI1HxgSZPN'),
    '81': ('model_4456.pth', '1JWRe3sIeHbDprHjzXitEkwypr5iZdO58'),
    '27': ('model_4590.pth', '1YJB01X7QmfB-hzVpUe9e_IQEiMvN3OIZ')
}


class MHFORMER:
    def __init__(self, model_version='351', batch_size='auto', device='cuda'):
        if model_version not in mhformer_model_map.keys():
            raise AssertionError(f'Only {", ".join(mhformer_model_map.keys())} models exist')

        model_name, cid = mhformer_model_map[model_version]
        self.model_path = os.path.join(os.path.dirname(__file__),
                                      f'MHFormer/models/{model_version}/{model_name}')

        if not os.path.exists(self.model_path):
            download_models(cid, self.model_path, f"MHformer {model_version}")

        self.layers, self.channel, self.d_hid, self.frames = 3, 512, 1024, int(model_version)
        self.n_joints, self.out_joints = 17, 17
        self.pad = self.frames // 2
        self.batch_size = self.get_batch_size() if batch_size == 'auto' else int(batch_size)
        self.device = device
        self.model = self.load_model()

    def __call__(self, keypoints_2d):

        keypoints_2d = coco2h36(keypoints_2d)

        seq_of_seq, len_seq = self.create_input(keypoints_2d)
        loader = DataLoader(seq_of_seq, batch_size=self.batch_size, shuffle=False)

        poses_3D = []
        for batch in tqdm(loader):
            batch_flip = self.flip_skeleton(batch.clone())
            batch = torch.concatenate((batch[None], batch_flip[None])).cuda()

            post_out = self.estimation(batch)
            poses_3D.append(post_out)

        poses_3D = np.concatenate(poses_3D)
        poses_3D = [poses_3D[len_seq[i]:len_seq[i + 1]] for i in range(len(len_seq) - 1)]
        clear()

        keypoints_3d = np.zeros((*keypoints_2d.shape[:-1], 3))
        for i in range(len(keypoints_2d)):
            keypoints_3d[i, (keypoints_2d[i] != 0).all(1).all(1)] = poses_3D[i]

        keypoints_3d *= [[[[-1, -1, 1]]]]
        return keypoints_3d

    @torch.no_grad()
    def estimation(self, batch):
        output_3D_non_flip = self.model(batch[0])[:, self.pad]
        output_3D_flip = self.model(batch[1])[:, self.pad]
        output_3D_flip = self.flip_skeleton(output_3D_flip)
        output_3D = (output_3D_non_flip + output_3D_flip) / 2
        output_3D[:, 0] = 0
        post_out = self.rotate_out(output_3D.cpu().detach().numpy())
        return post_out

    def create_input(self, keypoints):
        seq_of_seq = []
        len_seq = [0]
        for sequence in keypoints:
            sequence = sequence[(sequence != 0).all(1).all(1)]
            sequences = np.array(
                [self.normalize(self.make_sequence(sequence, i)) for i in range(len(sequence))])
            seq_of_seq.append(sequences.astype(np.float32))
            len_seq.append(len(sequences))

        seq_of_seq = np.concatenate(seq_of_seq)
        len_seq = np.array(len_seq).cumsum()
        return seq_of_seq, len_seq

    def make_sequence(self, keypoints, i):
        start = max(0, i - self.pad)
        end = min(i + self.pad, len(keypoints) - 1)
        sequence = keypoints[start:end + 1]

        left_pad, right_pad = 0, 0
        if sequence.shape[0] != self.frames:
            if i < self.pad:
                left_pad = self.pad - i
            if i > len(keypoints) - self.pad - 1:
                right_pad = i + self.pad - (len(keypoints) - 1)

            sequence = np.pad(sequence, ((left_pad, right_pad), (0, 0), (0, 0)), 'edge')
        return sequence

    @staticmethod
    def normalize(sequence):
        transp = sequence.transpose(1, 0, 2)
        shifted = transp - sequence.mean(axis=1)
        input_2D = shifted / (sequence.max(axis=1) - sequence.min(axis=1)).mean() / 2
        return input_2D.transpose(1, 0, 2)

    @staticmethod
    def flip_skeleton(sequence):
        joints_left = [4, 5, 6, 11, 12, 13]
        joints_right = [1, 2, 3, 14, 15, 16]
        if len(sequence.shape) == 4:
            sequence[:, :, :, 0] *= -1
            sequence[:, :, joints_left + joints_right] = sequence[:, :, joints_right + joints_left]
        elif len(sequence.shape) == 3:
            sequence[:, :, 0] *= -1
            sequence[:, joints_left + joints_right] = sequence[:, joints_right + joints_left]
        else:
            raise AssertionError('Len of shape must be equal 3 or 4')
        return sequence

    @staticmethod
    def get_batch_size():
        total = torch.cuda.get_device_properties(0).total_memory / 1024 ** 2
        return int(total // 30)

    @staticmethod
    def rotate_out(post_out):
        rot = np.float32([0.1407056450843811, -0.1500701755285263, -0.755240797996521, 0.6223280429840088])
        return camera_to_world(post_out, R=rot, t=0)

    def load_model(self):
        model = Model(self).to(torch.device(self.device))
        model_dict = model.state_dict()
        pre_dict = torch.load(self.model_path)
        for name, key in model_dict.items():
            model_dict[name] = pre_dict[name]
        model.load_state_dict(model_dict)
        model.eval()
        return model



# ------------------------------------------------------------------------------
# Adapted from https://github.com/leoxiaobin/deep-high-resolution-net.pytorch
# Original licence: Copyright (c) Microsoft, under the MIT License.
# ------------------------------------------------------------------------------

import math

import cv2
import numpy as np
import torch


def fliplr_joints(joints_3d, joints_3d_visible, img_width, flip_pairs):
    """Flip human joints horizontally.

    Note:
        - num_keypoints: K

    Args:
        joints_3d (np.ndarray([K, 3])): Coordinates of keypoints.
        joints_3d_visible (np.ndarray([K, 1])): Visibility of keypoints.
        img_width (int): Image width.
        flip_pairs (list[tuple]): Pairs of keypoints which are mirrored
            (for example, left ear and right ear).

    Returns:
        tuple: Flipped human joints.

        - joints_3d_flipped (np.ndarray([K, 3])): Flipped joints.
        - joints_3d_visible_flipped (np.ndarray([K, 1])): Joint visibility.
    """

    assert len(joints_3d) == len(joints_3d_visible)
    assert img_width > 0

    joints_3d_flipped = joints_3d.copy()
    joints_3d_visible_flipped = joints_3d_visible.copy()

    # Swap left-right parts
    for left, right in flip_pairs:
        joints_3d_flipped[left, :] = joints_3d[right, :]
        joints_3d_flipped[right, :] = joints_3d[left, :]

        joints_3d_visible_flipped[left, :] = joints_3d_visible[right, :]
        joints_3d_visible_flipped[right, :] = joints_3d_visible[left, :]

    # Flip horizontally
    joints_3d_flipped[:, 0] = img_width - 1 - joints_3d_flipped[:, 0]
    joints_3d_flipped = joints_3d_flipped * joints_3d_visible_flipped

    return joints_3d_flipped, joints_3d_visible_flipped


def fliplr_regression(regression,
                      flip_pairs,
                      center_mode='static',
                      center_x=0.5,
                      center_index=0):
    """Flip human joints horizontally.

    Note:
        - batch_size: N
        - num_keypoint: K

    Args:
        regression (np.ndarray([..., K, C])): Coordinates of keypoints, where K
            is the joint number and C is the dimension. Example shapes are:

            - [N, K, C]: a batch of keypoints where N is the batch size.
            - [N, T, K, C]: a batch of pose sequences, where T is the frame
                number.
        flip_pairs (list[tuple()]): Pairs of keypoints which are mirrored
            (for example, left ear -- right ear).
        center_mode (str): The mode to set the center location on the x-axis
            to flip around. Options are:

            - static: use a static x value (see center_x also)
            - root: use a root joint (see center_index also)
        center_x (float): Set the x-axis location of the flip center. Only used
            when center_mode=static.
        center_index (int): Set the index of the root joint, whose x location
            will be used as the flip center. Only used when center_mode=root.

    Returns:
        np.ndarray([..., K, C]): Flipped joints.
    """
    assert regression.ndim >= 2, f'Invalid pose shape {regression.shape}'

    allowed_center_mode = {'static', 'root'}
    assert center_mode in allowed_center_mode, 'Get invalid center_mode ' \
        f'{center_mode}, allowed choices are {allowed_center_mode}'

    if center_mode == 'static':
        x_c = center_x
    elif center_mode == 'root':
        assert regression.shape[-2] > center_index
        x_c = regression[..., center_index:center_index + 1, 0]

    regression_flipped = regression.copy()
    # Swap left-right parts
    for left, right in flip_pairs:
        regression_flipped[..., left, :] = regression[..., right, :]
        regression_flipped[..., right, :] = regression[..., left, :]

    # Flip horizontally
    regression_flipped[..., 0] = x_c * 2 - regression_flipped[..., 0]
    return regression_flipped


def flip_back(output_flipped, flip_pairs, target_type='GaussianHeatmap'):
    """Flip the flipped heatmaps back to the original form.

    Note:
        - batch_size: N
        - num_keypoints: K
        - heatmap height: H
        - heatmap width: W

    Args:
        output_flipped (np.ndarray[N, K, H, W]): The output heatmaps obtained
            from the flipped images.
        flip_pairs (list[tuple()): Pairs of keypoints which are mirrored
            (for example, left ear -- right ear).
        target_type (str): GaussianHeatmap or CombinedTarget

    Returns:
        np.ndarray: heatmaps that flipped back to the original image
    """
    assert output_flipped.ndim == 4, \
        'output_flipped should be [batch_size, num_keypoints, height, width]'
    shape_ori = output_flipped.shape
    channels = 1
    if target_type.lower() == 'CombinedTarget'.lower():
        channels = 3
        output_flipped[:, 1::3, ...] = -output_flipped[:, 1::3, ...]
    output_flipped = output_flipped.reshape(shape_ori[0], -1, channels,
                                            shape_ori[2], shape_ori[3])
    output_flipped_back = output_flipped.copy()

    # Swap left-right parts
    for left, right in flip_pairs:
        output_flipped_back[:, left, ...] = output_flipped[:, right, ...]
        output_flipped_back[:, right, ...] = output_flipped[:, left, ...]
    output_flipped_back = output_flipped_back.reshape(shape_ori)
    # Flip horizontally
    output_flipped_back = output_flipped_back[..., ::-1]
    return output_flipped_back


def transform_preds(coords, center, scale, output_size, use_udp=False):
    """Get final keypoint predictions from heatmaps and apply scaling and
    translation to map them back to the image.

    Note:
        num_keypoints: K

    Args:
        coords (np.ndarray[K, ndims]):

            * If ndims=2, corrds are predicted keypoint location.
            * If ndims=4, corrds are composed of (x, y, scores, tags)
            * If ndims=5, corrds are composed of (x, y, scores, tags,
              flipped_tags)

        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        use_udp (bool): Use unbiased data processing

    Returns:
        np.ndarray: Predicted coordinates in the images.
    """
    assert coords.shape[1] in (2, 4, 5)
    assert len(center) == 2
    assert len(scale) == 2
    assert len(output_size) == 2

    # Recover the scale which is normalized by a factor of 200.
    # scale = scale * 200.0

    if use_udp:
        scale_x = scale[0] / (output_size[0] - 1.0)
        scale_y = scale[1] / (output_size[1] - 1.0)
    else:
        scale_x = scale[0] / output_size[0]
        scale_y = scale[1] / output_size[1]

    target_coords = np.ones_like(coords)
    target_coords[:, 0] = coords[:, 0] * scale_x + center[0] - scale[0] * 0.5
    target_coords[:, 1] = coords[:, 1] * scale_y + center[1] - scale[1] * 0.5

    return target_coords


def get_affine_transform(center,
                         scale,
                         rot,
                         output_size,
                         shift=(0., 0.),
                         inv=False):
    """Get the affine transform matrix, given the center/scale/rot/output_size.

    Args:
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        rot (float): Rotation angle (degree).
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        shift (0-100%): Shift translation ratio wrt the width/height.
            Default (0., 0.).
        inv (bool): Option to inverse the affine transform direction.
            (inv=False: src->dst or inv=True: dst->src)

    Returns:
        np.ndarray: The transform matrix.
    """
    assert len(center) == 2
    assert len(scale) == 2
    assert len(output_size) == 2
    assert len(shift) == 2

    # pixel_std is 200.
    scale_tmp = scale * 200.0

    shift = np.array(shift)
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = rotate_point([0., src_w * -0.5], rot_rad)
    dst_dir = np.array([0., dst_w * -0.5])

    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    src[2, :] = _get_3rd_point(src[0, :], src[1, :])

    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def affine_transform(pt, trans_mat):
    """Apply an affine transformation to the points.

    Args:
        pt (np.ndarray): a 2 dimensional point to be transformed
        trans_mat (np.ndarray): 2x3 matrix of an affine transform

    Returns:
        np.ndarray: Transformed points.
    """
    assert len(pt) == 2
    new_pt = np.array(trans_mat) @ np.array([pt[0], pt[1], 1.])

    return new_pt


def _get_3rd_point(a, b):
    """To calculate the affine matrix, three pairs of points are required. This
    function is used to get the 3rd point, given 2D points a & b.

    The 3rd point is defined by rotating vector `a - b` by 90 degrees
    anticlockwise, using b as the rotation center.

    Args:
        a (np.ndarray): point(x,y)
        b (np.ndarray): point(x,y)

    Returns:
        np.ndarray: The 3rd point.
    """
    assert len(a) == 2
    assert len(b) == 2
    direction = a - b
    third_pt = b + np.array([-direction[1], direction[0]], dtype=np.float32)

    return third_pt


def rotate_point(pt, angle_rad):
    """Rotate a point by an angle.

    Args:
        pt (list[float]): 2 dimensional point to be rotated
        angle_rad (float): rotation angle by radian

    Returns:
        list[float]: Rotated point.
    """
    assert len(pt) == 2
    sn, cs = np.sin(angle_rad), np.cos(angle_rad)
    new_x = pt[0] * cs - pt[1] * sn
    new_y = pt[0] * sn + pt[1] * cs
    rotated_pt = [new_x, new_y]

    return rotated_pt


def get_warp_matrix(theta, size_input, size_dst, size_target):
    """Calculate the transformation matrix under the constraint of unbiased.
    Paper ref: Huang et al. The Devil is in the Details: Delving into Unbiased
    Data Processing for Human Pose Estimation (CVPR 2020).

    Args:
        theta (float): Rotation angle in degrees.
        size_input (np.ndarray): Size of input image [w, h].
        size_dst (np.ndarray): Size of output image [w, h].
        size_target (np.ndarray): Size of ROI in input plane [w, h].

    Returns:
        np.ndarray: A matrix for transformation.
    """
    theta = np.deg2rad(theta)
    matrix = np.zeros((2, 3), dtype=np.float32)
    scale_x = size_dst[0] / size_target[0]
    scale_y = size_dst[1] / size_target[1]
    matrix[0, 0] = math.cos(theta) * scale_x
    matrix[0, 1] = -math.sin(theta) * scale_x
    matrix[0, 2] = scale_x * (-0.5 * size_input[0] * math.cos(theta) +
                              0.5 * size_input[1] * math.sin(theta) +
                              0.5 * size_target[0])
    matrix[1, 0] = math.sin(theta) * scale_y
    matrix[1, 1] = math.cos(theta) * scale_y
    matrix[1, 2] = scale_y * (-0.5 * size_input[0] * math.sin(theta) -
                              0.5 * size_input[1] * math.cos(theta) +
                              0.5 * size_target[1])
    return matrix


def warp_affine_joints(joints, mat):
    """Apply affine transformation defined by the transform matrix on the
    joints.

    Args:
        joints (np.ndarray[..., 2]): Origin coordinate of joints.
        mat (np.ndarray[3, 2]): The affine matrix.

    Returns:
        np.ndarray[..., 2]: Result coordinate of joints.
    """
    joints = np.array(joints)
    shape = joints.shape
    joints = joints.reshape(-1, 2)
    return np.dot(
        np.concatenate((joints, joints[:, 0:1] * 0 + 1), axis=1),
        mat.T).reshape(shape)


def affine_transform_torch(pts, t):
    npts = pts.shape[0]
    pts_homo = torch.cat([pts, torch.ones(npts, 1, device=pts.device)], dim=1)
    out = torch.mm(t, torch.t(pts_homo))
    return torch.t(out[:2, :])



from .KeypointTracker import KeypointTracker
from .HumanVisualizer import Visualizer, scene_plotly
from .Networks import YOLOv7, VITPOSE, MHFORMER
from .HITrack import HITrack
from .constants import *
